"""
This file contains a library of functions related to power spectra estimates

"""

import os
import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter, fourier_gaussian
import scipy.stats as stats
from scipy.special import gamma
import astropy.units as u


#==================================================
# 2D pixel window function
#==================================================

def pixel_wf_mn(kx2d, ky2d, reso):
    """
    Compute the pixel window function for each k in 2D
    
    Parameters
    ----------
    - kx2d (2d np array): array of k along x (inverse unit of reso)
    - ky2d (2d np array): array of k along y (inverse unit of reso)
    - reso (float): the pixel size in inverse unit of k

    Outputs
    ----------
    - pix_mn (np array): the pixel window function for each k
    """

    term1 = np.sinc(kx2d*reso)
    term2 = np.sinc(ky2d*reso)
    pix_mn = term1 * term2
    
    return pix_mn


#==================================================
# 2D beam window function
#==================================================

def beam_wf_mn(kx2d, ky2d, FWHM):
    """
    Compute the beam window function for each k in 2D
    
    Parameters
    ----------
    - kx2d (2d np array): array of k along x (inverse unit of FWHM)
    - ky2d (2d np array): array of k along y (inverse unit of FWHM)
    - FWHM (float): the beam FWHM (same unit as 1/k)

    Outputs
    ----------
    - beam_mn (np array): the beam window function for each k in 2D
    """

    sigma2fwhm = 2 * np.sqrt(2*np.log(2))
    
    sigma = FWHM / sigma2fwhm
    k_0 = 1 / (np.sqrt(2) * np.pi * sigma)
    G_k = np.exp(-(kx2d**2 + ky2d**2) / k_0**2)
    
    return G_k


#==================================================
# 1D Power spectrum attenuation of a Gaussian beam
#==================================================

def beam_wf_pk(k, FWHM):
    """
    Compute the Gaussian beam window function power spectrum 
    (same as beam_wf_mn but with norm k as input, in 1D)
    
    Parameters
    ----------
    - k (np array): array of k, same unit as 1/FWHM
    - FWHM (float): Gaussian FWHM, same unit as 1/k

    Outputs
    ----------
    - G_k (np array): the power spectrum of a gaussian
    """
    
    sigma2fwhm = 2 * np.sqrt(2*np.log(2))

    sigma = FWHM / sigma2fwhm
    k_0 = 1 / (np.sqrt(2) * np.pi * sigma)
    G_k = np.exp(-k**2 / k_0**2)
    
    return G_k


#==================================================
# Apply NIKA-like transfer function
#==================================================
def apply_transfer_function(image, reso, beamFWHM, TF, apps_TF_LS=True, apps_beam=True):
    """
    Convolve SZ image with instrumental transfer function, decomposed in two: beam and processing.
    The beam is applied in Fourier to be better coherent for Pk estimates.
    
    Parameters
    ----------
    - image (np array): input raw image
    - reso (float): map resolution in arcsec
    - beamFWHM (float): the map beam FWHM in units of arcsec
    - TF (dict): dictionary with 'k' (unit homogeneous to 1/arcsec) 
    and 'TF' as keys, i.e. two np arrays containing the k and the 
    transfer function
    - apps_TF_LS (bool): set to true to apply the large scale TF filtering
    - apps_beam (bool): set to true to apply beam smoothing
        
    Outputs
    ----------
    - map_filt (2d np array): the convolved map
    """
    
    FT_map = np.fft.fft2(image)

    # Beam smoothing
    if apps_beam and beamFWHM>0:
        sigma2fwhm = 2 * np.sqrt(2*np.log(2))
        FT_map_sm = fourier_gaussian(FT_map, sigma=beamFWHM/sigma2fwhm/reso)
    else:
        FT_map_sm = FT_map*1 + 0
    
    # TF filtering
    if apps_TF_LS:
        Ny, Nx = image.shape
        k_x = np.fft.fftfreq(Nx, reso)
        k_y = np.fft.fftfreq(Ny, reso)
        k2d_x, k2d_y = np.meshgrid(k_x, k_y, indexing='xy')
        k2d_norm = np.sqrt(k2d_x**2 + k2d_y**2)
        k2d_norm_flat = k2d_norm.flatten()

        # interpolate by putting 1 outside the definition of the TF, i.e. no filtering, e.g. very small scale
        itpl = interp1d(TF['k'].to_value('arcsec-1'), TF['TF'], bounds_error=False, fill_value=(0,1))
        filtering_flat = itpl(k2d_norm_flat)
        filtering = np.reshape(filtering_flat, k2d_norm.shape)

        map_filt = np.real(np.fft.ifft2(FT_map_sm * filtering))
    else:
        map_filt = np.real(np.fft.ifft2(FT_map_sm))
        
    return map_filt


#==================================================
# Deconv NIKA-like transfer function
#==================================================

def deconv_transfer_function(image, reso, beamFWHM, TF, dec_TF_LS=True, dec_beam=True):
    """
    Deconvolve SZ image from instrumental transfer function.
    
    Parameters
    ----------
    - image (np array): input raw image
    - reso (float): map resolution in arcsec
    - beamFWHM (float): the map beam FWHM in units of arcsec
    - TF (dict): dictionary with 'k' (unit homogeneous to 1/arcsec) 
    and 'TF' as keys, i.e. two np arrays containing the k and the 
    transfer function
    - dec_TF_LS (bool): set to true to deconvol the large scale TF filtering
    - dec_beam (bool): set to true to deconvol beam smoothing
        
    Outputs
    ----------
    - map_deconv (2d np array): the deconvolved map
    """

    # FFT the map
    FT_map = np.fft.fft2(image)

    # Get the k arrays
    Ny, Nx = image.shape
    k_x = np.fft.fftfreq(Nx, reso)
    k_y = np.fft.fftfreq(Ny, reso)
    k2d_x, k2d_y= np.meshgrid(k_x, k_y, indexing='xy')
    k2d_norm = np.sqrt(k2d_x**2 + k2d_y**2)
    k2d_norm_flat = k2d_norm.flatten()

    #===== Beam smoothing
    if dec_beam:
        beam_mn = beam_wf_mn(k2d_x, k2d_y, beamFWHM)
        FT_map_B = FT_map / beam_mn

    else:
        FT_map_B = FT_map*1+0
            
    #===== TF filtering
    if dec_TF_LS:
        # interpolate by putting 1 outside the definition of the TF, i.e. no filtering, e.g. very small scale
        itpl = interp1d(TF['k'].to_value('arcsec-1'), TF['TF'], bounds_error=False, fill_value=(0,1))
        filtering_flat = itpl(k2d_norm_flat)
        filtering = np.reshape(filtering_flat, k2d_norm.shape)
    
        # Check undefined filtering
        condition = (filtering == 0)
        filtering[condition] = 1

        # Deconvolution in Fourier
        FT_map_TB = FT_map_B / filtering
        
        # Set the zero level depending on the TF at k=0
        FT_map_TB[condition] = 0 # keep the zero level untouched if 0 in TF

    else:
        FT_map_TB = FT_map_B*1+0

    #===== Back to real space
    map_deconv = np.real(np.fft.ifft2(FT_map_TB))

    return map_deconv


#==================================================
# Beam Power spectrum convolution
#==================================================

def apply_pk_beam(k, pk, beamFWHM1, beamFWHM2=None):
    """
    This function apply the power spectrum from beam 
    attenuation.
    
    Parameters
    ----------
    - k (nd array): k array in inverse scale unit
    - pk (nd array): power spectrum array
    - beamFWHM1 (quantity): the beam FWHM homogeneous to the inverse of k
    - beamFWHM2 (quantity): the second beam FWHM in case pk is a cross between 2 maps with different beam

    Outputs
    ----------
    - pk_conv (np array): the convolved pk

    """

    # Beam 1
    if beamFWHM1 > 0:
        Beam_k1 = beam_wf_pk(k, beamFWHM1)
    else:
        Beam_k1 = 1

    # Beam 2
    if beamFWHM2 != None:
        if beamFWHM2 > 0:
            Beam_k2 = beam_wf_pk(k, beamFWHM2)
        else:
            Beam_k2 = 1
    else:
        Beam_k2 = Beam_k1

    # Application
    pk_conv = pk*Beam_k1*Beam_k2

    return pk_conv


#==================================================
# Beam Power spectrum deconvolution
#==================================================

def deconv_pk_beam(k, pk, beamFWHM1, beamFWHM2=None):
    """
    This function corrects the power spectrum from beam 
    attenuation.
    
    Parameters
    ----------
    - k (nd array): k array in inverse scale unit
    - pk (nd array): power spectrum array
    - beamFWHM1 (quantity): the beam FWHM homogeneous to the inverse of k
    - beamFWHM2 (quantity): the second beam FWHM in case pk is a cross between 2 maps with different beam

    Outputs
    ----------
    - pk_deconv (np array): the deconvolved pk

    """

    # Beam 1
    if beamFWHM1 > 0:
        Beam_k1 = beam_wf_pk(k, beamFWHM1)
    else:
        Beam_k1 = 1
        
    # Beam 2
    if beamFWHM2 != None:
        if beamFWHM2 > 0:
            Beam_k2 = beam_wf_pk(k, beamFWHM2)
        else:
            Beam_k2 = 1
    else:
        Beam_k2 = Beam_k1

    # Application
    pk_deconv = pk/Beam_k1/Beam_k2

    return pk_deconv


#==================================================
# Transfer function power spectrum convolution
#==================================================

def apply_pk_transfer_function(k, pk, TF_k1, TF1,
                               TF_k2=None, TF2=None):
    """
    This function apply the power spectrum from transfer 
    function attenuation.
    
    Parameters
    ----------
    - k (nd array): k array in the same unit as in the TF_k
    - pk (nd array): power spectrum array
    - TF_k1 (nd array): k associated with the transfer function (same unit as k)
    - TF1 (nd array): filtering (same lenght as TF_k)
    - TF_k2 (nd array): same as TF_k1 in case of a second map with different TF
    - TF2 (nd array): same as TF1 in case of a second map with different TF

    Outputs
    ----------
    - pk_conv (np array): the convolved pk

    """

    # TF1
    itpl1 = interp1d(TF_k1, TF1, bounds_error=False, fill_value=(0,1))
    TF_kpc1 = itpl1(k)

    # TF2
    if TF_k2 is not None and TF2 is not None:
        itpl2 = interp1d(TF_k2, TF2, bounds_error=False, fill_value=(0,1))
        TF_kpc2 = itpl2(k)
    else:
        TF_kpc2 = TF_kpc1
    
    # Application
    pk_conv = pk*TF_kpc1*TF_kpc2
    
    return pk_conv


#==================================================
# Transfer function power spectrum deconvolution
#==================================================

def deconv_pk_transfer_function(k, pk, TF_k1, TF1,
                                TF_k2=None, TF2=None):
    """
    This function corrects the power spectrum from transfer 
    function attenuation.
    
    Parameters
    ----------
    - k (nd array): k array in the same unit as in the TF_k
    - pk (nd array): power spectrum array
    - TF_k1 (nd array): k associated with the transfer function (same unit as k)
    - TF1 (nd array): filtering (same lenght as TF_k)
    - TF_k2 (nd array): same as TF_k1 in case of a second map with different TF
    - TF2 (nd array): same as TF1 in case of a second map with different TF

    Outputs
    ----------
    - pk_deconv (np array): the deconvolved pk

    """

    # TF1
    itpl1 = interp1d(TF_k1, TF1, bounds_error=False, fill_value=(0,1))
    TF_kpc1 = itpl1(k)

    # TF2
    if TF_k2 is not None and TF2 is not None:
        itpl2 = interp1d(TF_k2, TF2, bounds_error=False, fill_value=(0,1))
        TF_kpc2 = itpl2(k)
    else:
        TF_kpc2 = TF_kpc1
    
    # Application
    pk_deconv = pk/TF_kpc1/TF_kpc2
    
    return pk_deconv


#==================================================
# Compute the matrix K used in Mbb calculations
#==================================================

def compute_Kmnmn(W):
    """
    Compute the matrix K_m,n,mp,np from Ponthieu et al. (2011)
    See Eq 10 and Appendix A.
    This gives accounts for masking/weighting effects in Fourier.
    Could certainly be optimized.

    Parameters
    ----------
    - W (complex 2d array): the FFT of the wight map

    Outputs
    ----------
    - K (np 4D array): the matrix K

    """
    
    Ny, Nx = W.shape
    K = np.zeros((Ny, Nx, Ny, Nx), dtype=complex)  # m, n, m1, n1
    
    for n1 in range(Nx):
        for m1 in range(Ny):
            for n in range(Nx):
                for m in range(Ny):
                    if m1 <= m and n1 <= n:
                        K[m, n, m1, n1] = W[m - m1, n - n1,]
                    elif m1 <= m and n1 > n:
                        K[m, n, m1, n1] = W[m - m1, Nx + n - n1]
                    elif m1 > m and n1 <= n:
                        K[m, n,  m1, n1] = W[Ny + m - m1, n - n1]
                    elif m1 > m and n1 > n:
                        K[m, n, m1, n1] = W[Ny + m - m1, Nx + n - n1]
    
    return K


#==================================================
# Compute the bining operators R,Q from POKER
#==================================================

def compute_RQ(k2d, kedge, beta=0):
    """
    Compute the binning operators R and Q from Ponthieu et al. (2011)
    See Eq 11,12

    Parameters
    ----------
    - k2d (complex 2d array): the norm of k on the 2d grid
    - kedge (1D array): the edges of the bins
    - beta (float): bin k^beta Pk

    Outputs
    ----------
    - R (np array): the bining operator
    - Q (np array): the inverse bining operator

    """
       
    Ny, Nx = k2d.shape
    Nb = len(kedge)-1
    R = np.zeros((Nb, Ny, Nx))
    Q = np.zeros((Nb, Ny, Nx))
    
    for b in range(Nb):
        wbin = (k2d >= kedge[b]) * (k2d < kedge[b+1])
        sigma = np.sum(wbin) # Number of kmn falling in bin
        for n in range(Nx):
            for m in range(Ny):
                cond = (kedge[b] <= k2d[m,n]) and (kedge[b+1] > k2d[m,n])
                if cond:
                    if beta == 0:
                        R[b, m,n] = 1.0 / sigma 
                        Q[b, m,n] = 1.0
                    else:
                        R[b, m,n] = k2d[m,n]**beta / sigma
                        Q[b, m,n] = 1/k2d[m,n]**beta
                else:
                    R[b, m,n] = 0.0
                    Q[b, m,n] = 0.0
                    
    return R, Q


#==================================================
# Compute the bin-to-bin mixing matrix
#==================================================

def compute_Mbb(R, K, Q):
    """
    Compute the bin-to-bin mixing matrix Mbb from Ponthieu et al. (2011)
    See Eq 16

    Parameters
    ----------
    - R (nd array): the bining operator
    - K (nd array): the K_m,n,mp,np array
    - Q (nd array): the inverse bining operator

    Outputs
    ----------
    - Mbb (2d array): the bin to bin mixing matrix

    """
    
    Nb, Ny, Nx = R.shape
    Mbb = np.zeros((Nb, Nb))
    for ib in range(Nb):
        for jb in range(Nb):            
                sum_KQ  = np.sum(np.abs(K)**2 * (Q[jb, :,:])[np.newaxis, np.newaxis, :, :], axis=(2,3))
                sum_RKQ = np.sum(R[ib, :,:] * sum_KQ)
                Mbb[ib, jb] = sum_RKQ
                
    return Mbb/(Nx*Ny)**2


#==================================================
# Apply K matrix to 2d power spectra
#==================================================

def multiply_Kmnmn(K, T):
    """
    Application of K onto 2D power spectra T.
    Eq. A.5 from Ponthieu+2011

    Parameters
    ----------
    - K (complex 2d array): output of 'compute_Kmnmn'
    - T (2d array): the input model for the power spectrum on a 2d grid

    Outputs
    ----------
    - KxT (np 2D array): the power spectrum in 2D convolved with K

    """

    Ny, Nx = T.shape
    KxT = np.zeros((Ny, Nx), dtype=complex)  # m,n
    
    for n in range(Nx):
        for m in range(Ny):
            KxT[m, n] = np.sum(K[m, n, :, :] * T) 
        
    return KxT / Ny/Nx


def multiply_Kmnmn_bis(K, T):
    """
    Same as multiply_Kmnmn but avoid loop.
    Application of K onto 2D power spectra T.
    Eq. A.5 from Ponthieu+2011

    Parameters
    ----------
    - K (complex 2d array): output of 'compute_Kmnmn'
    - T (2d array): the input model for the power spectrum on a 2d grid

    Outputs
    ----------
    - KxT (np 2D array): the power spectrum in 2D convolved with K

    """

    Ny, Nx = T.shape
    KxT = np.zeros((Ny, Nx), dtype=complex)  # m,n
    T_reshaped = np.tile(T, (Ny, Nx, 1, 1))
    KxT = np.sum(K * T_reshaped, axis=(2, 3))
        
    return KxT / Nx/Ny


#==================================================
# Measure the 3D power spectrum naively
#==================================================

def convert_pkln_to_pkgauss(Pk_r, reso_x, reso_y, reso_z):
    """
    Let Pk_r be the power spectrum of the simulated lognormal field r
    and Pk_s be the power spectrum of the gaussian field s that is
    used to generate the lognormal field as
    field_ln = exp(s)
    This function convert Pk_r to the Pk_s
    Note that the field mean is zero here and should be defined a 
    posteriori in real space (see e.g. get_pressure_cube_fluctuation in 
    model_mock.py).
    See M. Greiner and T. A. Enßlin, A&A 574, A86 (2015)
        
    Parameters
    ----------
    - Pk_r (3d np array): lognormal field powerspectrum
    - reso_x (float): the resolution along the x axis
    - reso_y (float): the resolution along the y axis
    - reso_z (float): the resolution along the z axis

    Outputs
    ----------
    - Pk_s (np array): Pk followed by the gaussian field 
    """
    
    # Get sampling info
    Nz, Ny, Nx = Pk_r.shape
    Vx = reso_x * reso_y * reso_z
    Vk = 1 / (Nx*Ny*Nz*Vx)

    k_x = np.fft.fftfreq(Nx, reso_x) # 1/kpc
    k_y = np.fft.fftfreq(Ny, reso_y)
    k_z = np.fft.fftfreq(Nz, reso_z)
    k3d_z, k3d_y, k3d_x = np.meshgrid(k_z, k_y, k_x, indexing='ij')
    knorm = (k3d_z**2 + k3d_y**2 + k3d_x**2)**0.5

    # Apply the convertion
    I    = np.abs(np.fft.fftn(Pk_r)*Vk + 1)    
    Pk_s = np.abs(np.fft.ifftn(np.log(I))*Vx*Nx*Ny*Nz)
    Pk_s[knorm == 0] = 0
    
    return Pk_s


#==================================================
# Measure the 3D power spectrum naively
#==================================================

def extract_pk3d(cube, proj_reso, los_reso,
                 cube2=None,
                 Nbin=100, scalebin='lin',
                 kmin=None, kmax=None, kedges=None,
                 statistic='mean'):
    """
    Measure the power spectrum in 3 dimensions in k bins.
    The unit of the k array is the inverse of the resolution
    The unit of Pk is that of the resolution element cubed.
    
    Parameters
    ----------
    - cube (np array): 3d data cube as nd array
    - proj_reso (float): the resolution along the projected direction
    - los_reso (float) the resolution along the line of sight
    - cube2 (np array): same as cube. Use to do cross spectra
    - Nbin (int): the number of bin for output Pk. If Nbin is None, then no 
    binning is done
    - scalebin (str): lin or log, the way the Pk is binned along k
    - kmin/max (float): the min and max k to use in defining the bins
    - kedges (1d np array): directly provide the bin edges in a Nbin+1 array (in this case,
    - statistic (str or function): the statistics to be used in binned_statistic
        
    Outputs
    ----------
    - kvals (np array): values of k binned
    - Pk_bins (np array): values of amplitude binned
    """
    
    # Get the number of pixels
    Nz, Ny, Nx = cube.shape

    # Define the k_i and k_norm
    k_x = np.fft.fftfreq(Nx, proj_reso)
    k_y = np.fft.fftfreq(Ny, proj_reso)
    k_z = np.fft.fftfreq(Nz, los_reso)
    k3d_z, k3d_y, k3d_x = np.meshgrid(k_z, k_y, k_x, indexing='ij')
    k3d_norm = np.sqrt(k3d_x**2 + k3d_y**2 + k3d_z**2)
    
    # Compute the Pk cube
    if cube2 is None:
        fourier_cube = np.fft.fftn(cube)
        fourier_pk   = np.abs(fourier_cube)**2 # img unit squared
    else:
        fourier_c1 = np.fft.fftn(cube)
        fourier_c2 = np.fft.fftn(cube2)
        fourier_pk = np.real(fourier_c1*np.conjugate(fourier_c2)+np.conjugate(fourier_c1)*fourier_c2)/2

    # Get the flattened k, Pk
    knrm = k3d_norm.flatten()
    fourier_pk = fourier_pk.flatten()

    #----- Case of binning
    if Nbin is not None:
        # Define the bins
        if kedges is None:
            if kmin is None:
                kmin_sampling = np.amin(k3d_norm[k3d_norm > 0])
            else:
                kmin_sampling = kmin
                
            if kmax is None:
                kmax_sampling = np.amax(k3d_norm)
            else:
                kmax_sampling = kmax
                
            if scalebin == 'lin':
                kbins = np.linspace(kmin_sampling, kmax_sampling, Nbin+1)
            elif scalebin == 'log':
                kbins = np.logspace(np.log10(kmin_sampling), np.log10(kmax_sampling), Nbin+1)
            else:
                raise ValueError("Only lin or log scales are allowed. Here scalebin="+scalebin)
            
        else:
            kbins = kedges
            
        kvals = 0.5 * (kbins[1:] + kbins[:-1])
        
        # Bin the Pk
        Pk_bins, _, _ = stats.binned_statistic(knrm, fourier_pk,statistic=statistic, bins=kbins)
        Pk_bins *= (proj_reso*proj_reso*los_reso) / (Nx*Ny*Nz)

    #----- Case unbinned
    else:
        kvals = knrm
        Pk_bins = fourier_pk * (proj_reso*proj_reso*los_reso) / (Nx*Ny*Nz)

    return kvals, Pk_bins


#==================================================
# Measure the 3D power spectrum naively
#==================================================

def extract_pk2d(image, proj_reso,
                 image2=None,
                 Nbin=100, scalebin='lin', kmin=None, kmax=None, kedges=None,
                 statistic='mean'):
    """
    Measure the power spectrum in 2 dimensions in k bins.
    The unit of the k array is the inverse of the resolution
    The unit of Pk is that of the resolution element squared.
    
    Parameters
    ----------
    - image (np array): 2d data image as nd array
    - proj_reso (float): the resolution along the projected direction (any unit, e.g. kpc, arcsec)
    - image2 (np array): same as image. Use to do cross spectra
    - Nbin (int): the number of bin for output Pk. If Nbin is None, then no 
    binning is done
    - scalebin (str): lin or log, the way the Pk is binned along k
    - kmin/max (float): the min and max k to use in defining the bins
    - kedges (1d np array): directly provide the bin edges in a Nbin+1 array (in this case,
      Nbin, kmin/kmax and scalebin are ignored)
    - statistic (str or function): the statistics to be used in binned_statistic
        
    Outputs
    ----------
    - kvals (np array): values of k binned
    - Pk_bins (np array): values of amplitude binned
    """
    
    # Get the number of pixels
    Ny, Nx = image.shape

    # Define the k_i and k_norm
    k_x = np.fft.fftfreq(Nx, proj_reso)
    k_y = np.fft.fftfreq(Ny, proj_reso)
    k2d_x, k2d_y = np.meshgrid(k_x, k_y, indexing='xy')
    k2d_norm = np.sqrt(k2d_x**2 + k2d_y**2)
    
    # Compute the Pk amplitide
    if image2 is None:
        fourier_img = np.fft.fftn(image)
        fourier_pk = np.abs(fourier_img)**2 # img unit squared
    else:
        fourier_img1 = np.fft.fftn(image)
        fourier_img2 = np.fft.fftn(image2)
        fourier_pk = np.real(fourier_img1*np.conjugate(fourier_img2)+np.conjugate(fourier_img1)*fourier_img2)/2

    # Get the flattened k, Pk
    knrm = k2d_norm.flatten()
    fourier_pk = fourier_pk.flatten()

    #----- Case of binning
    if Nbin is not None:
        # Define the bins
        if kedges is None:
            if kmin is None:
                kmin_sampling = np.amin(k2d_norm[k2d_norm > 0])
            else:
                kmin_sampling = kmin
                
            if kmax is None:
                kmax_sampling = np.amax(k2d_norm)
            else:
                kmax_sampling = kmax
                
            if scalebin == 'lin':
                kbins = np.linspace(kmin_sampling, kmax_sampling, Nbin+1)
            elif scalebin == 'log':
                kbins = np.logspace(np.log10(kmin_sampling), np.log10(kmax_sampling), Nbin+1)
            else:
                raise ValueError("Only lin or log scales are allowed. Here scalebin="+scalebin)
            
        else:
            kbins = kedges
            
        kvals = 0.5 * (kbins[1:] + kbins[:-1])
        
        # Bin the Pk
        Pk_bins, _, _ = stats.binned_statistic(knrm, fourier_pk, statistic=statistic, bins=kbins)
        Pk_bins *= (proj_reso*proj_reso) / (Nx*Ny) # img unit squared x reso unit squared

    #----- Case unbinned
    else:
        kvals = knrm
        Pk_bins = fourier_pk * (proj_reso*proj_reso) / (Nx*Ny)

    return kvals, Pk_bins


#==================================================
# Measure the 3D power spectrum naively
#==================================================

def extract_pk2d_arevalo(image, proj_reso,
                         kctr=None,
                         Nbin=100, scalebin='lin', kmin=None, kmax=None, kedges=None,
                         epsilon=1e-3, mask=None,
                         unbias_apply=False, unbias_slope=None, unbias_beamFWHM=0):
    """
    Implement the method of Arevalo et al. (2012) to compute the power spectrum.
    Warning, the method is biased for steep spectra, which is the case 
    near the beam cutoff. The unbiais keyword attempt to unbias the spectrum
    from itself.

    The unit of output k is the same as the inverse of the resolution, and that of Pk 
    is the same as resolution^2.
    
    See also Romero et al. (2023) and its erratum for the bias estimate due to the PSF.

    Parameters
    ----------
    - image (np array): 2d data cube as nd array
    - proj_reso (float): the resolution along the projected direction (any unit, e.g. kpc, arcsec)
    - kctr (nd array): the k values at which to compute Pk. If given, Nbin, scalebin, kmin/max,
      and kedges are irrelevant.
    - Nbin (int): the number of bin for output Pk
    - scalebin (str): lin or log, the way the Pk is binned along k
    - kmin/max (float): the min and max k to use in defining the bins
    - kedges (1d np array): directly provide the bin edges in a Nbin+1 array (in this case,
      Nbin, kmin/kmax and scalebin are ignored)
    - epsilon (float): parameter of the sigma calculation
    - mask (nd array): same size as image, but with 0 or 1 to mask bad pixels
    - unbias_apply (bool): set to true to unbias the spectrum given the estimated power law bias
    using the first guess measured spectrum as the input
    - unbias_slope (float or array with lenght Nbin): the slope of the underlying true spectrum 
    needed for debiasing
    - unbias_beamFWHM (float): the Gaussian beam FWHM (same unit as proj_reso)

    Outputs
    ----------
    - k2d (np array): values of k2d
    - Pk (np array): values of spectrum at k2d
    """

    #----- Possibility to define the bins in different ways        
    if kctr is None:    # The user does not gives the central k values
        # First need the k grid
        k_x = np.fft.fftfreq(image.shape[1], proj_reso)
        k_y = np.fft.fftfreq(image.shape[0], proj_reso)
        k2d_x, k2d_y = np.meshgrid(k_x, k_y, indexing='xy')
        k2d_norm = np.sqrt(k2d_x**2 + k2d_y**2)

        # Define the kmin/max from the grid or the user
        if kmin is None:
            kmin_sampling = np.amin(k2d_norm[k2d_norm > 0])
        else:
            kmin_sampling = kmin
        if kmax is None:
            kmax_sampling = np.amax(k2d_norm)
        else:
            kmax_sampling = kmax

        # get the bins
        if scalebin is 'lin':
            kbins = np.linspace(kmin_sampling, kmax_sampling, Nbin+1)
        elif scalebin is 'log':
            kbins = np.logspace(np.log10(kmin_sampling), np.log10(kmax_sampling), Nbin+1)
        else:
            raise ValueError("Only lin or log scales are allowed")

        # Defines the bin accounting for possible user defined bins
        if kedges is None:
            kbins = kbins
        else:
            kbins = kedges

        # Get the central values
        k2d = 0.5 * (kbins[1:] + kbins[:-1])

    else:           # The user directly gives the central k values
        k2d = kctr

    #----- Make sure the image is masked
    if mask is not None:
        image = image*mask
        
    #----- Compute the spectrum
    sigma2fwhm = 2 * np.sqrt(2*np.log(2))
    sigma_list = 1 / (k2d * np.sqrt(2 * np.pi**2))

    Pk=[]
    for sigma in sigma_list:
        k_i = 1 / (sigma * np.sqrt(2 * np.pi**2))
        sigma1 = sigma / np.sqrt(1+epsilon)
        sigma2 = sigma * np.sqrt(1+epsilon)
        img_conv1 = apply_transfer_function(image, proj_reso, sigma2fwhm*sigma1, 0, apps_TF_LS=False)
        img_conv2 = apply_transfer_function(image, proj_reso, sigma2fwhm*sigma2, 0, apps_TF_LS=False)

        if mask is not None:
            mask_conv1 = apply_transfer_function(mask, proj_reso, sigma2fwhm*sigma1, 0, apps_TF_LS=False)
            mask_conv2 = apply_transfer_function(mask, proj_reso, sigma2fwhm*sigma2, 0, apps_TF_LS=False)
            wbad1 = (mask_conv1 == 0) 
            wbad2 = (mask_conv2 == 0)
            mask_conv1[wbad1] = np.nan
            mask_conv2[wbad2] = np.nan
            img_conv = (img_conv1/mask_conv1 - img_conv2/mask_conv2) * mask
            img_conv[wbad1] = 0
            img_conv[wbad2] = 0
            var = np.sum(img_conv**2) / img_conv.size * (mask.size / np.sum(mask))
        else:
            img_conv = (img_conv1 - img_conv2)
            var = np.sum(img_conv**2) / img_conv.size
            
        spec_i = var / epsilon**2 / np.pi / k_i**2
        Pk.append(spec_i)
        
    Pk = np.array(Pk)

    #----- Attempt to unbias the spectrum
    if unbias_apply:
        # Compute the slope from the spectrum itself if not provided (i.e. do what you can)
        if unbias_slope is None:
            der = -np.gradient(np.log10(Pk), np.log10(k2d))
        else:
            der = unbias_slope
        biasPL = 2**(der/2) * gamma(3 - der/2) / gamma(3)
        
        # Compute the beam smoothing bias correction
        if unbias_beamFWHM > 0:
            sigma_i = unbias_beamFWHM/sigma2fwhm
            k_i = 1 / (np.sqrt(2) * np.pi * sigma_i)
            x_i = k_i / k2d
            biasB = (1 + 1/x_i**2)**(-3+der/2)
        else:
            biasB = 1
        
        Pk = Pk/(biasPL*biasB)

    return k2d, Pk






"""
This file deals with the model parametrization

"""

import astropy.units as u
import numpy as np
from scipy import interpolate
import warnings
from astropy import constants as const
from scipy.interpolate import interp1d

from minot.ClusterTools import cluster_global
from minot.ClusterTools import cluster_profile


#==================================================
# Library class
#==================================================

class ModelLibrary(object):
    """ ModelAdmin class
    This class searves as a parser to the main Model class, to
    include the subclass ModelAdmin in this other file. All the definitions of the
    model parameters should be here.

    Profile models are now:  ['GNFW', 'SVM', 'beta', 'doublebeta']

    Attributes
    ----------
    The attributes are the same as the Model class, see model_main.py

    Methods
    ----------
    - set_pressure_profile_universal_param(self, pressure_model='P13UPP'): set the gas pressure
    profile parameters to the universal value from different results
    - _validate_profile_model_parameters(self, inpar, unit): dedicated to check and validate the parameters
    of profile models
    - _get_generic_profile(self, radius, model, derivative=False): get any profile base on model type

    """

    #==================================================
    # Set a given pressure UPP profile
    #==================================================

    def set_pressure_profile_universal_param(self, pressure_model='P13UPP'):
        """
        Set the parameters of the pressure profile:
        P0, c500 (and r_p given R500), gamma, alpha, beta
        
        Parameters
        ----------
        - pressure_model (str): available models are
            'A10UPP' (Universal Pressure Profile from Arnaud et al. 2010)
            'A10CC' (Cool-Core Profile from Arnaud et al. 2010)
            'A10MD' (Morphologically-Disturbed Profile from Arnaud et al. 2010)
            'P13UPP' (Planck Intermediate Paper V (2013) Universal Pressure Profile)
            'G19UPP' (Ghirardini et al 2019, all clusters)
            'G19CC' (Ghirardini et al 2019, cool-core clusters)
            'G19MD' (Ghirardini et al 2019, morphologically disturbed clusters)
        """

        # Arnaud et al. (2010) : Universal Pressure Profile parameters
        if pressure_model == 'A10UPP':
            if not self._silent: print('Setting gNFW Arnaud et al. (2010) UPP.')
            pppar = [8.403, 1.177, 0.3081, 1.0510, 5.4905]

        # Arnaud et al. (2010) : Cool-Core clusters parameters
        elif pressure_model == 'A10CC':
            if not self._silent: print('Setting gNFW Arnaud et al. (2010) cool-core.')
            pppar = [3.249, 1.128, 0.7736, 1.2223, 5.4905]

        # Arnaud et al. (2010) : Morphologically-Disturbed clusters parameters
        elif pressure_model == 'A10MD':
            if not self._silent: print('Setting gNFW Arnaud et al. (2010) morphologically disturbed.')
            pppar = [3.202, 1.083, 0.3798, 1.4063, 5.4905]

        # Planck Intermediate Paper V (2013) : Universal Pressure Profile parameters
        elif pressure_model == 'P13UPP':
            if not self._silent: print('Setting gNFW Planck coll. (2013) UPP.')
            pppar = [6.410, 1.810, 0.3100, 1.3300, 4.1300]

        # Ghirardini (2019) : Universal Pressure Profile parameters
        elif pressure_model == 'G19UPP':
            if not self._silent: print('Setting gNFW Ghirardini (2019) UPP.')
            pppar = [5.68, 1.49, 0.43, 1.33, 4.40]

        # Ghirardini (2019) : Universal Pressure Profile parameters
        elif pressure_model == 'G19CC':
            if not self._silent: print('Setting gNFW Ghirardini (2019) CC.')
            pppar = [6.03, 1.68, 0.51, 1.33, 4.37]

            # Ghirardini (2019) : Universal Pressure Profile parameters
        elif pressure_model == 'G19MD':
            if not self._silent: print('Setting gNFW Ghirardini (2019) MD.')
            pppar = [7.96, 1.79, 0.29, 1.33, 4.05]

        # No other profiles available
        else:
            raise ValueError('P(r) model not available. Use A10UPP,A10CC,A10MD,P13UPP,G19UPP,G19CC,G19MD.')

        # Compute the normalization
        if pressure_model in ['A10UPP', 'A10CC', 'A10MD', 'P13UPP']:
            mu_g,mu_e,mu_p,mu_a = cluster_global.mean_molecular_weight(Y=self._helium_mass_fraction,
                                                                       Z=self._metallicity_sol*self._abundance)
            try:
                fb = self._cosmo.Ob0/self._cosmo.Om0
            except:
                fb = 0.16
            Pnorm = cluster_global.gNFW_normalization(self._redshift,
                                                      self._M500.to_value('Msun'),
                                                      cosmo=self._cosmo,
                                                      mu=mu_g, mue=mu_e, fb=fb)
        if pressure_model in ['G19UPP', 'G19CC', 'G19MD']:
            E_z   = self._cosmo.efunc(self._redshift)
            h70   = self._cosmo.H0.value/70.0
            mu_g,mu_e,mu_p,mu_a = cluster_global.mean_molecular_weight(Y=self._helium_mass_fraction,
                                                                       Z=self._metallicity_sol*self._abundance)
            Pnorm = 3.426*1e-3 * (self._M500.to_value('Msun')*h70/1e15)**(2.0/3) * E_z**(8.0/3)
            try:
                fb = self._cosmo.Ob0/self._cosmo.Om0
            except:
                fb = 0.16
            Pnorm = Pnorm * (fb/0.16) * (mu_g/0.6) * (mu_e/1.14)
        
        # Set the parameters accordingly
        self._model_pressure_profile = {"name": 'GNFW',
                                        "P_0" : pppar[0]*Pnorm*u.Unit('keV cm-3'),
                                        "c500": pppar[1],
                                        "r_p" : self._R500/pppar[1],
                                        "a":pppar[3],
                                        "b":pppar[4],
                                        "c":pppar[2]}
    
    
    #==================================================
    # Validate profile model parameters
    #==================================================
    
    def _validate_model_profile_parameters(self, inpar, unit):
        """
        Check the profile parameters.
        
        Parameters
        ----------
        - inpar (dict): a dictionary containing the input parameters
        - unit (str): contain the unit of the profile, e.g. keV/cm-3 for pressure
        
        Outputs
        ----------
        - outpar (dict): a dictionary with output parameters

        """

        # List of available authorized models
        model_list = ['NFW', 'GNFW', 'SVM', 'beta', 'doublebeta', 'User']
        
        # Deal with unit
        if unit == '' or unit == None:
            hasunit = False
        else:
            hasunit = True

        # Check that the input is a dictionary
        if type(inpar) != dict :
            raise TypeError("The model should be a dictionary containing the name key and relevant parameters")
        
        # Check that input contains a name
        if 'name' not in list(inpar.keys()) :
            raise ValueError("The model dictionary should contain a 'name' field")
            
        # Check that the name is in the acceptable name list
        if not inpar['name'] in model_list:
            print('The profile model can be:')
            print(model_list)
            raise ValueError("The requested model is not available")

        #---------- Deal with the case of NFW
        if inpar['name'] == 'NFW':
            # Check the content of the dictionary
            cond1a = 'r_s' in list(inpar.keys()) and 'rho_0' in list(inpar.keys())
            cond1b = not('M' in list(inpar.keys()) or 'c' in list(inpar.keys()) or 'delta' in list(inpar.keys()))
            cond2a = 'M' in list(inpar.keys()) and 'c' in list(inpar.keys()) and 'delta' in list(inpar.keys())
            cond2b = not('rho_0' in list(inpar.keys()) or 'r_s' in list(inpar.keys()))

            if (cond1a and cond1b):
                case1 = True
            else:
                case1 = False
            if (cond2a and cond2b):
                case2 = True
            else:
                case2 = False
            
            if not ((cond1a and cond1b) or (cond2a and cond2b)):
                raise ValueError("The NFW model should contain: {'rho_0', 'r_s'} or {'M', 'c', 'delta'} .")

            # Case of generic profile with rho0 and rs
            if case1:
                # Check units and values
                if hasunit:
                    try:
                        test = inpar['rho_0'].to(unit)
                    except:
                        raise TypeError("rho_0 should be homogeneous to "+unit)
                
                if inpar['rho_0'] < 0:
                    raise ValueError("rho_0 should be >=0")
                
                try:
                    test = inpar['r_s'].to('kpc')
                except:
                    raise TypeError("r_s should be homogeneous to kpc")
                    
                if inpar['r_s'] <= 0:
                    raise ValueError("r_s should be >0")
                
                # All good at this stage, setting parameters
                if hasunit:
                    rho0 = inpar['rho_0'].to(unit)
                else:
                    rho0 = inpar['rho_0']*u.adu

                r_s = inpar['r_s'].to('kpc')

            # Case of physical profile with mass and concentration
            if case2:
                # Check units and values
                if hasunit:
                    try:
                        test = (u.Unit(unit)).to('Msun kpc-3')
                    except:
                        raise TypeError("The unit should be left empty or homogeneous to Msun/kpc^3 for a physically defined NFW")

                try:
                    test = inpar['M'].to('Msun')
                except:
                    raise TypeError("M should be homogeneous to Msun")
                    
                if inpar['M'] <= 0:
                    raise ValueError("M should be >0")
                
                if inpar['c'] <= 0:
                    raise ValueError("c should be >0")

                if inpar['delta'] < 100 or inpar['delta'] > 5000:
                    raise ValueError("delta is restricted to 100 < delta < 5000 to avoid numerical faillures")

                # Compute r_s and rho_0
                R_delta = cluster_global.Mdelta_to_Rdelta(inpar['M'].to_value('Msun'), self._redshift,
                                                          delta=inpar['delta'], cosmo=self._cosmo)
                r_s = (R_delta / inpar['c']) * u.kpc
                overdens = cluster_global.concentration_to_deltac_NFW(inpar['c'], Delta=inpar['delta'])
                rho0 = overdens * self._cosmo.critical_density(self._redshift).to('Msun kpc-3')

            # Set the parameters
            outpar = {"name": 'NFW',
                      "rho_0" : rho0,
                      "r_s"   : r_s}
            
        #---------- Deal with the case of GNFW
        elif inpar['name'] == 'GNFW':
            # Check the content of the dictionary
            cond1 = 'P_0' in list(inpar.keys()) and 'a' in list(inpar.keys()) and 'b' in list(inpar.keys()) and 'c' in list(inpar.keys())
            cond2 = 'c500' in list(inpar.keys()) or 'r_p' in list(inpar.keys())
            cond3 = not('c500' in list(inpar.keys()) and 'r_p' in list(inpar.keys()))

            if not (cond1 and cond2 and cond3):
                raise ValueError("The GNFW model should contain: {'P_0','c500' or 'r_p' (not both),'a','b','c'}.")
            
            # Check units and values
            if hasunit:
                try:
                    test = inpar['P_0'].to(unit)
                except:
                    raise TypeError("P_0 should be homogeneous to "+unit)

            if inpar['P_0'] < 0:
                raise ValueError("P_0 should be >=0")

            if 'r_p' in list(inpar.keys()):
                try:
                    test = inpar['r_p'].to('kpc')
                except:
                    raise TypeError("r_p should be homogeneous to kpc")
                
                if inpar['r_p'] <= 0:
                    raise ValueError("r_p should be >0")

            if 'c500' in list(inpar.keys()):
                if inpar['c500'] <=0:
                    raise ValueError("c500 should be > 0")

            # All good at this stage, setting parameters
            if 'c500' in list(inpar.keys()):
                c500 = inpar['c500']
                r_p = self._R500/inpar['c500']
            if 'r_p' in list(inpar.keys()):
                c500 = (self._R500/inpar['r_p']).to_value('')
                r_p = inpar['r_p']
            if hasunit:
                P0 = inpar['P_0'].to(unit)
            else:
                P0 = inpar['P_0']*u.adu
                
            outpar = {"name": 'GNFW',
                      "P_0" : P0,
                      "c500": c500,
                      "r_p" : r_p.to('kpc'),
                      "a"   : inpar['a'],
                      "b"   : inpar['b'],
                      "c"   : inpar['c']}
            
        #---------- Deal with the case of SVM
        if inpar['name'] == 'SVM':
            # Check the content of the dictionary
            cond1 = 'n_0' in list(inpar.keys()) and 'r_c' in list(inpar.keys()) and 'beta' in list(inpar.keys())
            cond2 = 'r_s' in list(inpar.keys()) and 'alpha' in list(inpar.keys()) and 'epsilon' in list(inpar.keys()) and 'gamma' in list(inpar.keys())
            if not (cond1 and cond2):
                raise ValueError("The SVM model should contain: {'n_0','beta','r_c','r_s', 'alpha', 'gamma', 'epsilon'}.")
 
            # Check units
            if hasunit:
                try:
                    test = inpar['n_0'].to(unit)
                except:
                    raise TypeError("n_0 should be homogeneous to "+unit)
            try:
                test = inpar['r_c'].to('kpc')
            except:
                raise TypeError("r_c should be homogeneous to kpc")
            try:
                test = inpar['r_s'].to('kpc')
            except:
                raise TypeError("r_s should be homogeneous to kpc")
            
            # Check values
            if inpar['n_0'] < 0:
                raise ValueError("n_0 should be >= 0")
            if inpar['r_c'] <= 0:
                raise ValueError("r_c should be larger than 0")
            if inpar['r_s'] <= 0:
                raise ValueError("r_s should be larger than 0")

            if hasunit:
                n0 = inpar['n_0'].to(unit)
            else:
                n0 = inpar['n_0']*u.adu
            
            # All good at this stage, setting parameters
            outpar = {"name"   : 'SVM',
                      "n_0"    : n0,
                      "r_c"    : inpar['r_c'].to('kpc'),
                      "r_s"    : inpar['r_s'].to('kpc'),
                      "alpha"  : inpar['alpha'],
                      "beta"   : inpar['beta'],
                      "gamma"  : inpar['gamma'],
                      "epsilon": inpar['epsilon']}
            
        #---------- Deal with the case of beta
        if inpar['name'] == 'beta':
            # Check the content of the dictionary
            cond1 = 'n_0' in list(inpar.keys()) and 'r_c' in list(inpar.keys()) and 'beta' in list(inpar.keys())
            if not cond1:
                raise ValueError("The beta model should contain: {'n_0','beta','r_c'}.")

            # Check units
            if hasunit:
                try:
                    test = inpar['n_0'].to(unit)
                except:
                    raise TypeError("n_0 should be homogeneous to "+unit)
            try:
                test = inpar['r_c'].to('kpc')
            except:
                raise TypeError("r_c should be homogeneous to kpc")

            # Check values
            if inpar['n_0'] < 0:
                raise ValueError("n_0 should be >= 0")
            if inpar['r_c'] <= 0:
                raise ValueError("r_c should be larger than 0")

            if hasunit:
                n0 = inpar['n_0'].to(unit)
            else:
                n0 = inpar['n_0']*u.adu
                
            # All good at this stage, setting parameters
            outpar = {"name"   : 'beta',
                      "n_0"    : n0,
                      "r_c"    : inpar['r_c'].to('kpc'),
                      "beta"   : inpar['beta']}
            
        #---------- Deal with the case of doublebeta
        if inpar['name'] == 'doublebeta':
            # Check the content of the dictionary
            cond1 = 'n_01' in list(inpar.keys()) and 'r_c1' in list(inpar.keys()) and 'beta1' in list(inpar.keys())
            cond2 = 'n_02' in list(inpar.keys()) and 'r_c2' in list(inpar.keys()) and 'beta2' in list(inpar.keys())
            if not (cond1 and cond2):
                raise ValueError("The double beta model should contain: {'n_01','beta1','r_c1','n_02','beta2','r_c2'}.")

            # Check units
            if hasunit:
                try:
                    test = inpar['n_01'].to(unit)
                except:
                    raise TypeError("n_01 should be homogeneous to "+unit)
                try:
                    test = inpar['n_02'].to(unit)
                except:
                    raise TypeError("n_02 should be homogeneous to "+unit)
            try:
                test = inpar['r_c1'].to('kpc')
            except:
                raise TypeError("r_c1 should be homogeneous to kpc")
            try:
                test = inpar['r_c2'].to('kpc')
            except:
                raise TypeError("r_c2 should be homogeneous to kpc")

            # Check values
            if inpar['n_01'] < 0:
                raise ValueError("n_01 should be >= 0")
            if inpar['r_c1'] <= 0:
                raise ValueError("r_c1 should be larger than 0")
            if inpar['n_02'] < 0:
                raise ValueError("n_02 should be >= 0")
            if inpar['r_c2'] <= 0:
                raise ValueError("r_c2 should be larger than 0")

            if hasunit:
                n01 = inpar['n_01'].to(unit)
                n02 = inpar['n_02'].to(unit)
            else:
                n01 = inpar['n_01']*u.adu
                n02 = inpar['n_02']*u.adu
            
            # All good at this stage, setting parameters
            outpar = {"name"  : 'doublebeta',
                      "n_01"  : n01,
                      "r_c1"  : inpar['r_c1'].to('kpc'),
                      "beta1" : inpar['beta1'],
                      "n_02"  : n02,
                      "r_c2"  : inpar['r_c2'].to('kpc'),
                      "beta2" : inpar['beta2']}

        #---------- Deal with the case of User
        if inpar['name'] == 'User':
            # Check the content of the dictionary
            cond1 = 'radius' in list(inpar.keys()) and 'profile' in list(inpar.keys())
            if not cond1:
                raise ValueError("The User model should contain: {'radius','profile'}.")

            # Check units
            if hasunit:
                try:
                    test = inpar['profile'].to(unit)
                except:
                    raise TypeError("profile should be homogeneous to "+unit)
            try:
                test = inpar['radius'].to('kpc')
            except:
                raise TypeError("radius should be homogeneous to kpc")

            # Check values
            if np.amin(inpar['radius']) < 0:
                raise ValueError("radius should be >= 0")
            if np.amin(inpar['profile']) < 0:
                raise ValueError("profile should be larger >= 0")

            if hasunit:
                prof = inpar['profile'].to(unit)
            else:
                prof = inpar['profile']*u.adu

            # All good at this stage, setting parameters
            outpar = {"name"    : 'User',
                      "radius"  : inpar['radius'].to('kpc'),
                      "profile" : prof}
            
        return outpar


    #==================================================
    # Validate fluctuation model parameters
    #==================================================
    
    def _validate_model_fluctuation_parameters(self, inpar):
        """
        Check the pressure fluctuation model parameters.
        
        Parameters
        ----------
        - inpar (dict): a dictionary containing the input parameters
        - unit (str): contain the unit of the fluctuation, if any
        
        Outputs
        ----------
        - outpar (dict): a dictionary with output parameters

        """

        # List of available authorized models
        model_list = ['CutoffPowerLaw']
        stats_list = ['gaussian', 'lognormal']
        
        # Check that the input is a dictionary
        if type(inpar) != dict :
            raise TypeError("The model should be a dictionary containing the name key and relevant parameters")
        
        # Check that input contains a name
        if 'name' not in list(inpar.keys()) :
            raise ValueError("The model dictionary should contain a 'name' field")
            
        # Check that the name is in the acceptable name list
        if not inpar['name'] in model_list:
            print('The fluctuation model can be:')
            print(model_list)
            raise ValueError("The requested model is not available")

        # Check that input contains a statistics
        if 'statistics' not in list(inpar.keys()) :
            raise ValueError("The model dictionary should contain a 'statistics' field")
            
        # Check that the statistics is acceptable
        if not inpar['statistics'] in stats_list:
            print('The fluctuation "statistics" can be:')
            print(stats_list)
            raise ValueError("The requested statistics are not available")

        #---------- Deal with the case of CutoffPowerLaw
        if inpar['name'] == 'CutoffPowerLaw':
            # Check the content of the dictionary
            cond1 = 'Norm' in list(inpar.keys()) and 'slope' in list(inpar.keys())
            cond2 = 'Linj' in list(inpar.keys()) and 'Ldis' in list(inpar.keys())
            if not (cond1 and cond2):
                raise ValueError("The CutoffPowerLaw model should contain: {'Norm','slope','Linj','Ldis'}.")

            # Check units
            try:
                test = inpar['Linj'].to('kpc')
            except:
                raise TypeError("Linj should be homogeneous to kpc")
            try:
                test = inpar['Ldis'].to('kpc')
            except:
                raise TypeError("Ldis should be homogeneous to kpc")
            if type(inpar['Norm']) == u.quantity.Quantity:
                raise TypeError("Norm is the rms of the relative fluctuation, should be dimenssionless")
            if type(inpar['slope']) == u.quantity.Quantity:
                raise TypeError("slope should be dimenssionless")

            # Check values
            if inpar['Norm'] < 0:
                raise ValueError("Norm should be >= 0")
            if inpar['Linj'] < 0:
                raise ValueError("Linj should be >= 0")
            if inpar['Ldis'] < 0:
                raise ValueError("Ldis should be >= 0")

            # All good at this stage, setting parameters
            outpar = {"name" : 'CutoffPowerLaw',
                      "statistics": inpar['statistics'],
                      "Norm" : inpar['Norm'],
                      "slope": inpar['slope'],
                      "Linj" : inpar['Linj'].to('kpc'),
                      "Ldis" : inpar['Ldis'].to('kpc')}

        return outpar


    #==================================================
    # Get the generic model profile
    #==================================================

    def _get_generic_profile(self, radius, model, derivative=False):
        """
        Get the generic profile profile.
        
        Parameters
        ----------
        - radius (quantity) : the physical 3d radius in units homogeneous to kpc, as a 1d array
        - model (dict): dictionary containing the model parameters
        - derivative (bool): to get the derivative of the profile
        
        Outputs
        ----------
        - radius (quantity): the 3d radius in unit of kpc
        - p_r (quantity): the profile

        """

        model_list = ['NFW', 'GNFW', 'SVM', 'beta', 'doublebeta', 'User']

        if not model['name'] in model_list:
            print('The profile model can :')
            print(model_list)
            raise ValueError("The requested model has not been implemented")

        r3d_kpc = radius.to_value('kpc')

        #---------- Case of GNFW profile
        if model['name'] == 'NFW':
            unit = model["rho_0"].unit
            
            rho0 = model["rho_0"].to_value(unit)
            rs = model["r_s"].to_value('kpc')

            if derivative:
                prof_r = cluster_profile.NFW_model_derivative(r3d_kpc, rho0, rs) * unit*u.Unit('kpc-1')
            else:
                prof_r = cluster_profile.NFW_model(r3d_kpc, rho0, rs)*unit

        #---------- Case of GNFW profile
        elif model['name'] == 'GNFW':
            unit = model["P_0"].unit
            
            P0 = model["P_0"].to_value(unit)
            rp = model["r_p"].to_value('kpc')
            a  = model["a"]
            b  = model["b"]
            c  = model["c"]

            if derivative:
                prof_r = cluster_profile.gNFW_model_derivative(r3d_kpc, P0, rp,
                                                               slope_a=a,
                                                               slope_b=b,
                                                               slope_c=c) * unit*u.Unit('kpc-1')
            else:
                prof_r = cluster_profile.gNFW_model(r3d_kpc, P0, rp, slope_a=a, slope_b=b, slope_c=c)*unit

        #---------- Case of SVM model
        elif model['name'] == 'SVM':
            unit = model["n_0"].unit
            
            n0      = model["n_0"].to_value(unit)
            rc      = model["r_c"].to_value('kpc')
            rs      = model["r_s"].to_value('kpc')
            alpha   = model["alpha"]
            beta    = model["beta"]
            gamma   = model["gamma"]
            epsilon = model["epsilon"]

            if derivative:
                prof_r = cluster_profile.svm_model_derivative(r3d_kpc, n0, rc, beta,
                                                              rs, gamma, epsilon, alpha) * unit*u.Unit('kpc-1')
            else:
                prof_r = cluster_profile.svm_model(r3d_kpc, n0, rc, beta, rs, gamma, epsilon, alpha)*unit

        #---------- beta model
        elif model['name'] == 'beta':
            unit = model["n_0"].unit

            n0      = model["n_0"].to_value(unit)
            rc      = model["r_c"].to_value('kpc')
            beta    = model["beta"]

            if derivative:
                prof_r = cluster_profile.beta_model_derivative(r3d_kpc, n0, rc, beta) * unit*u.Unit('kpc-1')
            else:
                prof_r = cluster_profile.beta_model(r3d_kpc, n0, rc, beta)*unit
            
        #---------- double beta model
        elif model['name'] == 'doublebeta':
            unit1 = model["n_01"].unit
            unit2 = model["n_02"].unit

            n01      = model["n_01"].to_value(unit1)
            rc1      = model["r_c1"].to_value('kpc')
            beta1    = model["beta1"]
            n02      = model["n_02"].to_value(unit2)
            rc2      = model["r_c2"].to_value('kpc')
            beta2    = model["beta2"]

            if derivative:
                prof_r1 = cluster_profile.beta_model_derivative(r3d_kpc, n01, rc1, beta1) * unit1*u.Unit('kpc-1')
                prof_r2 = cluster_profile.beta_model_derivative(r3d_kpc, n02, rc2, beta2) * unit2*u.Unit('kpc-1')
                prof_r = prof_r1 + prof_r2
            else:
                prof_r1 = cluster_profile.beta_model(r3d_kpc, n01, rc1, beta1)*unit1
                prof_r2 = cluster_profile.beta_model(r3d_kpc, n02, rc2, beta2)*unit2
                prof_r = prof_r1 + prof_r2

        #---------- User model
        elif model['name'] == 'User':
            user_r = model["radius"].to_value('kpc')
            user_p = model["profile"].value

            # Warning
            if np.amin(user_r) > np.amin(r3d_kpc) or np.amax(user_r) < np.amax(r3d_kpc):
                if self._silent == False:
                    print('WARNING: User model interpolated beyond the provided range!')
            
            # Case of derivative needed
            if derivative:
                # Compute the derivative
                user_der = np.gradient(user_p, user_r)
                f = interpolate.interp1d(user_r, user_der, kind='linear', fill_value='extrapolate')
                prof_r = f(r3d_kpc)
                
                # Add the unit
                prof_r = prof_r*u.Unit('kpc-1')*model["profile"].unit

            # Standard case
            else:
                # log-log interpolation of the user profile
                with warnings.catch_warnings(): # Warning raised in the case of log(0)
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    f = interpolate.interp1d(np.log10(user_r),
                                             np.log10(user_p), kind='linear', fill_value='extrapolate')
                    pitpl_r = 10**f(np.log10(r3d_kpc))

                # Correct for nan (correspond to user_p == 0 in log)
                pitpl_r[np.isnan(pitpl_r)] = 0
                
                # Correct for negative value
                pitpl_r[pitpl_r<0] = 0

                # Add the unit
                prof_r = pitpl_r*model["profile"].unit

        #---------- Otherwise nothing is done
        else :
            if not self._silent: print('The requested model has not been implemented.')

        return prof_r




"""
This file contains the Model class. It is dedicated to the construction of a
Model object, definined by its physical properties and with associated methods
to compute derived properties or observables.

"""

#==================================================
# Requested imports
#==================================================

import os
import numpy as np
import astropy.units as u
from astropy.io import fits
import astropy.cosmology
from astropy.coordinates import SkyCoord
from astropy import constants as const
from astropy.wcs import WCS
from minot.ClusterTools import cluster_global
import pprint
import dill

from pitszi                  import title
from pitszi.model_library    import ModelLibrary
from pitszi.model_sampling   import ModelSampling
from pitszi.model_mock       import ModelMock


#==================================================
# Cluster Model class
#==================================================

class Model(ModelLibrary, ModelSampling, ModelMock):
    """ Model class.
    This class defines a model object. In addition to basic properties such as 
    mass and redshift, it includes the required physical properties for the modeling
    (pressure profile, pressure fluctuations).

    Attributes
    ----------  
    - silent (bool): print information if False, or not otherwise.
    - output_dir (str): directory where to output data files and plots.
    - cosmology (astropy.cosmology): background cosmological model. Can only be set
    when creating the Cluster object.
    - name (str): the name of the cluster
    - coord (SkyCoord object): the coordinate of the cluster.
    - redshift (float): redshift of the cluster center. Changing the redshift 
    on the fly propagate to cluster properties.
    - D_ang (quantity): can be access but not set directly. The redshift+cosmo
    defines this.
    - D_lum (quantity) : can be access but not set directly. The redshift+cosmo
    defines this.
    - M500 (quantity) : the mass enclosed within R500.
    - R500 (quantity) : the radius in which the density is 500 times the critical 
    density at the cluster redshift
    - theta500 (quantity): the angle corresponding to R500.
    - R_truncation (quantity): the radius at which the cluster stops (similar as virial radius)
    - theta_truncation (quantity): the angle corresponding to R_truncation.
    - Rmin (quantity): minimal scale considered numerically
    - helium_mass_fraction (float): the helium mass fraction of the gas (==Yp~0.25 in BBN)
    - metallicity_sol (float): the metallicity (default is Zprotosun == 0.0153)
    - abundance (float): the abundance (default is 0.3) in unit of Zprotosun
    - model_pressure_profile (dict): the model used for the thermal gas electron pressure 
    profile. It contains the name of the model and the associated model parameters. 
    - model_pressure_fluctuation (dict): the model used for the thermal gas electron pressure 
    fluctuations. It contains the name of the model and the associated model parameters. 
    - map_center (SkyCoord object): the map center coordinates.
    - map_reso (quantity): the map pixel size, homogeneous to degrees.
    - map_fov (list of quantity):  the map field of view as [FoV_x, FoV_y], homogeneous to deg.
    - map_header (standard header): this allows the user to provide a header directly.
    In this case, the map coordinates, field of view and resolution will be extracted 
    from the header and the projection can be arbitrary. If the header is not provided,
    then the projection will be standard RA-DEC tan projection.

    Methods
    ----------  
    Methods are split into the following files
    - title (title page)
    - model_library (model library related functions)
    - model_sampling (grid sampling related functions)
    - model_mock (mock generation related functions)

    ToDo
    ----------  
    - Improve the 3D pressure profile cube to avoid numercial issues near the center
    - Implement the possibility to have radial dependant fluctuations

    """

    #==================================================
    # Initialize the cluster object
    #==================================================

    def __init__(self,
                 name='Cluster',
                 RA=0.0*u.deg, Dec=0.0*u.deg,
                 redshift=0.5,
                 M500=1e15*u.Unit('Msun'),
                 cosmology=astropy.cosmology.Planck15,
                 silent=False,
                 output_dir='./pitszi_output',
    ):
        """
        Initialize the cluster object. Several attributes of the class cannot 
        be defined externally because of intrications between parameters. For 
        instance, the cosmology cannot be changed on the fly because this would 
        mess up the internal consistency.
        
        Parameters
        ----------
        - name (str): cluster name 
        - RA, Dec (quantity): coordinates or the cluster in equatorial frame
        - redshift (float) : the cluster center cosmological redshift
        - M500 (quantity): the cluster mass 
        - cosmology (astropy.cosmology): the name of the cosmology to use.
        - silent (bool): set to true in order not to print informations when running 
        - output_dir (str): where to save outputs
        
        """
        
        #---------- Print the code header at launch
        if not silent:
            title.show_model()
            
        #---------- Admin
        self._silent     = silent
        self._output_dir = output_dir

        #---------- Check that the cosmology is indeed a cosmology object
        if hasattr(cosmology, 'h') and hasattr(cosmology, 'Om0'):
            self._cosmo = cosmology
        else:
            raise TypeError("Input cosmology must be an instance of astropy.cosmology")
        
        #---------- Global properties
        self._name     = name
        self._coord    = SkyCoord(RA, Dec, frame="icrs")
        self._redshift = redshift
        self._D_ang    = self._cosmo.angular_diameter_distance(self._redshift)
        self._D_lum    = self._cosmo.luminosity_distance(self._redshift)
        self._M500     = M500
        self._R500     = cluster_global.Mdelta_to_Rdelta(self._M500.to_value('Msun'),
                                                         self._redshift, delta=500,
                                                         cosmo=self._cosmo)*u.kpc
        self._theta500 = ((self._R500 / self._D_ang).to('') * u.rad).to('deg')

        # Initial x, y, z (l.o.s.) axes correspond to min, intermediate, maj axis
        # Euler angles 1, 2, 3 are about z axis, x axis, and z axis
        # see https://en.wikipedia.org/wiki/Euler_angles, https://arxiv.org/pdf/1702.00795.pdf
        self._triaxiality = {'min_to_maj_axis_ratio':1, 'int_to_maj_axis_ratio':1,
                             'euler_angle1':0*u.deg, 'euler_angle2':0*u.deg, 'euler_angle3':0*u.deg}
        
        #---------- Cluster boundary
        self._R_truncation     = 3*self._R500
        self._theta_truncation = ((self._R_truncation / self._D_ang).to('') * u.rad).to('deg')
        self._Rmin = 10.0*u.kpc

        #---------- ICM composition (default: protosolar from Lodders et al 2009: arxiv0901.1149)
        self._helium_mass_fraction = 0.2735
        self._metallicity_sol = 0.0153
        self._abundance = 0.3
        
        #---------- P3D physical properties
        self._model_pressure_profile = 1
        self.set_pressure_profile_universal_param(pressure_model='P13UPP')
        
        dPpar = self._validate_model_fluctuation_parameters({"name":'CutoffPowerLaw',
                                                             "statistics":'gaussian', # or lognormal
                                                             "Norm":0.25,
                                                             "slope":-11/3.0, 
                                                             "Linj":1*u.Mpc,
                                                             "Ldis":1*u.kpc})
        self._model_pressure_fluctuation = dPpar
        
        #---------- Sampling
        self._map_center = SkyCoord(RA, Dec, frame="icrs")
        self._map_reso   = 5*u.arcsec
        self._map_fov    = [5.0, 5.0]*u.arcmin
        self._map_header = None
        self._los_reso   = 20*u.kpc
        self._los_size   = 2 * 2*u.Mpc


    #==================================================
    # Print parameters
    #==================================================
    
    def print_param(self):
        """
        Print the current parameters describing the inference.
        
        Parameters
        ----------
        
        Outputs
        ----------
        The parameters are printed in the terminal
        
        """

        print('=====================================================')
        print('=============== Current Model() state ===============')
        print('=====================================================')
        
        pp = pprint.PrettyPrinter(indent=4)
        
        par = self.__dict__
        keys = list(par.keys())
        
        for k in range(len(keys)):
            print(('--- '+(keys[k])[1:]))
            print(('    '+str(par[keys[k]])))
            print(('    '+str(type(par[keys[k]]))+''))
        print('=====================================================')


    #==================================================
    # Save parameters
    #==================================================
    
    def save_model(self):
        """
        Save the current model object.
        
        Parameters
        ----------
            
        Outputs
        ----------
        The parameters are saved in the output directory

        """

        # Create the output directory if needed
        if not os.path.exists(self.output_dir): os.mkdir(self.output_dir)

        # Save
        with open(self.output_dir+'/model_parameters.pkl', 'wb') as pfile:
            #pickle.dump(self.__dict__, pfile, pickle.HIGHEST_PROTOCOL)
            dill.dump(self.__dict__, pfile)

        # Text file for user
        par = self.__dict__
        keys = list(par.keys())
        with open(self.output_dir+'/model_parameters.txt', 'w') as txtfile:
            for k in range(len(keys)):
                txtfile.write('--- '+(keys[k])+'\n')
                txtfile.write('    '+str(par[keys[k]])+'\n')
                txtfile.write('    '+str(type(par[keys[k]]))+'\n')

                
    #==================================================
    # Load parameters
    #==================================================
    
    def load_model(self, param_file):
        """
        Read a given parameter file to re-initialize the model object.
        
        Parameters
        ----------
        param_file (str): the parameter file to be read
            
        Outputs
        ----------
            
        """

        with open(param_file, 'rb') as pfile:
            par = dill.load(pfile)
            
        self.__dict__ = par
        
        
    #==================================================
    # Get the hidden variable
    #==================================================
    
    #========== Admin
    @property
    def silent(self):
        return self._silent
    
    @property
    def output_dir(self):
        return self._output_dir
    
    #========== Cosmology
    @property
    def cosmo(self):
        return self._cosmo
    
    #========== Global properties
    @property
    def name(self):
        return self._name
    
    @property
    def coord(self):
        return self._coord
    
    @property
    def redshift(self):
        return self._redshift
    
    @property
    def D_ang(self):
        return self._D_ang
    
    @property
    def D_lum(self):
        return self._D_lum
    
    @property
    def M500(self):
        return self._M500
    
    @property
    def R500(self):
        return self._R500
    
    @property
    def theta500(self):
        return self._theta500

    @property
    def triaxiality(self):
        return self._triaxiality
 
    #========== Cluster boundary
    @property
    def R_truncation(self):
        return self._R_truncation
    
    @property
    def theta_truncation(self):
        return self._theta_truncation
    
    @property
    def Rmin(self):
        return self._Rmin

    #========== ICM composition
    @property
    def helium_mass_fraction(self):
        return self._helium_mass_fraction

    @property
    def abundance(self):
        return self._abundance

    @property
    def metallicity_sol(self):
        return self._metallicity_sol
    
    #========== ICM physics
    @property
    def model_pressure_profile(self):
        return self._model_pressure_profile
    
    @property
    def model_pressure_fluctuation(self):
        return self._model_pressure_fluctuation
    
    #========== Maps parameters
    @property
    def map_center(self):
        return self.get_map_center()
    
    @property
    def map_reso(self):
        return self.get_map_reso()
    
    @property
    def map_fov(self):
        return self.get_map_fov()
    
    @property
    def map_header(self):
        return self.get_map_header()

    @property
    def los_reso(self):
        return self._los_reso

    @property
    def los_size(self):
        return self._los_size

    
    #==================================================
    # Defines how the user can pass arguments and interconnections
    #==================================================
    
    #========== Admin
    @silent.setter
    def silent(self, value):
        # Check value and set
        if type(value) == bool:
            self._silent = value
        else:
            raise TypeError("The silent parameter should be a boolean.")

        # Information
        if not self._silent: print("Setting silent value")

    @output_dir.setter
    def output_dir(self, value):
        # Check value and set
        if type(value) == str:
            self._output_dir = value
        else:
            raise TypeError("The output_dir should be a string.")
        
        # Information
        if not self._silent: print("Setting output_dir value")

    #========== Cosmology
    @cosmo.setter
    def cosmo(self, value):
        message = ("The cosmology can only be set when defining the cluster object, "
                   "as clust = Cluster(cosmology=astropy.cosmology.YourCosmology).  "
                   "Doing nothing.                                                  ")
        if not self._silent: print(message)
    

    #========== Global properties
    @name.setter
    def name(self, value):
        # Check value and set
        if type(value) == str:
            self._name = value
        else:
            raise TypeError("The name should be a string.")

        # Information
        if not self._silent: print("Setting name value")

    @coord.setter
    def coord(self, value):
        # Case value is a SkyCoord object, nothing to be done
        if type(value) == astropy.coordinates.sky_coordinate.SkyCoord:
            self._coord = value

        # Case value is standard coordinates
        elif type(value) == dict:
            
            # It is not possible to have both RA-Dec and Glat-Glon, or just RA and not Dec, etc
            cond1 = 'RA'  in list(value.keys()) and 'Glat' in list(value.keys())
            cond2 = 'RA'  in list(value.keys()) and 'Glon' in list(value.keys())
            cond3 = 'Dec' in list(value.keys()) and 'Glat' in list(value.keys())
            cond4 = 'Dec' in list(value.keys()) and 'Glon' in list(value.keys())
            if cond1 or cond2 or cond3 or cond4:
                raise TypeError("The coordinates can be a coord object, or a {'RA','Dec'} or {'Glon', 'Glat'} dictionary.")
            
            # Case where RA-Dec is used
            if 'RA' in list(value.keys()) and 'Dec' in list(value.keys()):
                self._coord = SkyCoord(value['RA'], value['Dec'], frame="icrs")

            # Case where Glon-Glat is used
            elif 'Glon' in list(value.keys()) and 'Glat' in list(value.keys()):
                self._coord = SkyCoord(value['Glon'], value['Glat'], frame="galactic")

            # Otherwise, not appropriate value
            else:
                err_message = ("The coordinates can be a coord object, "
                               "a {'RA','Dec'} dictionary, or a {'Glon', 'Glat'} dictionary.")
                raise TypeError(err_message)

        # Case value is not accepted
        else:
            raise TypeError("The coordinates can be a coord object, a {'RA','Dec'} dictionary, or a {'Glon', 'Glat'} dictionary.")

        # Information
        if not self._silent: print("Setting coord value")

    @redshift.setter
    def redshift(self, value):
        # check type
        if type(value) != float and type(value) != int and type(value) != np.float64:
            raise TypeError("The redshift should be a int or a float.")
        
        # value check
        if value < 0 :
            raise ValueError("The redshift should be larger or equal to 0.")

        # Setting parameters
        self._redshift = value
        self._D_ang = self._cosmo.angular_diameter_distance(self._redshift)
        self._D_lum = self._cosmo.luminosity_distance(self._redshift)
        self._R500  = cluster_global.Mdelta_to_Rdelta(self._M500.to_value('Msun'),
                                                      self._redshift, delta=500, cosmo=self._cosmo)*u.kpc            
        self._theta500 = ((self._R500 / self._D_ang).to('') * u.rad).to('deg')
        self._theta_truncation = ((self._R_truncation / self._D_ang).to('') * u.rad).to('deg')
        
        # Information
        if not self._silent: print("Setting redshift value")
        if not self._silent: print("Setting: D_ang, D_lum, R500, theta500, theta_truncation ; Fixing: cosmo.")
        
    @D_ang.setter
    def D_ang(self, value):
        if not self._silent: print("The angular diameter distance cannot be set directly, the redshift has to be used instead.")
        if not self._silent: print("Doing nothing.                                                                            ")
        
    @D_lum.setter
    def D_lum(self, value):
        if not self._silent: print("The luminosity distance cannot be set directly, the redshift has to be used instead.")
        if not self._silent: print("Doing nothing.                                                                      ")
        
    @M500.setter
    def M500(self, value):
        # Type check
        try:
            test = value.to('Msun')
        except:
            raise TypeError("The mass M500 should be a quantity homogeneous to Msun.")

        # Value check
        if value <= 0 :
            raise ValueError("Mass M500 should be larger than 0")
        if value.to_value('Msun') < 1e10 :
            print("Warning! Your are setting the mass to a tiny value (i.e. not the cluster regime). This may lead to issues")
        
        # Setting parameters
        self._M500 = value
        self._R500 = cluster_global.Mdelta_to_Rdelta(self._M500.to_value('Msun'),
                                                     self._redshift, delta=500, cosmo=self._cosmo)*u.kpc
        self._theta500 = ((self._R500 / self._D_ang).to('') * u.rad).to('deg')
        
        # Information
        if not self._silent: print("Setting M500 value")
        if not self._silent: print("Setting: R500, theta500 ; Fixing: redshift, cosmo, D_ang")
        
    @R500.setter
    def R500(self, value):
        # Check type
        try:
            test = value.to('kpc')
        except:
            raise TypeError("The radius R500 should be a quantity homogeneous to kpc.")

        # check value
        if value < 0 :
            raise ValueError("Radius R500 should be larger than 0")

        # Setting parameter
        self._R500 = value
        self._theta500 = ((self._R500 / self._D_ang).to('') * u.rad).to('deg')
        self._M500 = cluster_global.Rdelta_to_Mdelta(self._R500.to_value('kpc'),
                                                     self._redshift, delta=500, cosmo=self._cosmo)*u.Msun
        
        # Information
        if not self._silent: print("Setting R500 value")
        if not self._silent: print("Setting: theta500, M500 ; Fixing: redshift, cosmo, D_ang")
        
    @theta500.setter
    def theta500(self, value):
        # Check type
        try:
            test = value.to('deg')
        except:
            raise TypeError("The angle theta500 should be a quantity homogeneous to deg.")
        
        # Check value
        if value <= 0 :
            raise ValueError("Angle theta500 should be larger than 0")

        # Setting parameters
        self._theta500 = value
        self._R500 = value.to_value('rad')*self._D_ang
        self._M500 = cluster_global.Rdelta_to_Mdelta(self._R500.to_value('kpc'),
                                                     self._redshift, delta=500, cosmo=self._cosmo)*u.Msun
        
        # Information
        if not self._silent: print("Setting theta500 value")
        if not self._silent: print("Setting: R500, M500 ; Fixing: redshift, cosmo, D_ang")

    @triaxiality.setter
    def triaxiality(self, value):
        # Check type
        if type(value) != dict :
            raise TypeError("The triaxiality should be a dictionary with relevant keys")

        cond1 = 'min_to_maj_axis_ratio' in list(value.keys()) and 'int_to_maj_axis_ratio' in list(value.keys())
        cond2 = 'euler_angle1' in list(value.keys()) and 'euler_angle2' in list(value.keys()) and 'euler_angle3' in list(value.keys())
        if not (cond1 and cond2):
            raise ValueError("The triaxiality should contain {min/int}_to_maj_axis_ratio and euler_angle{1,2,3} parameters")

        # Test units
        try:
            test = value['euler_angle1'].to('deg')
            test = value['euler_angle2'].to('deg')
            test = value['euler_angle3'].to('deg')
        except:
            raise ValueError("The euler_angle_{1,2,3} should have unit homogeneous to deg")

        # Test values
        cond1 = value['min_to_maj_axis_ratio'] <= 0 
        cond2 = value['min_to_maj_axis_ratio'] >  1
        cond3 = value['int_to_maj_axis_ratio'] <= 0 
        cond4 = value['int_to_maj_axis_ratio'] >  1
        cond5 = value['min_to_maj_axis_ratio'] > value['int_to_maj_axis_ratio']
        if cond1 or cond2 or cond3 or cond4 or cond5:
            raise ValueError("By definition, we should have 0 < min_to_maj_axis_ratio <= int_to_maj_axis_ratio <= 1")
        
        # Setting parameters
        self._triaxiality = {'min_to_maj_axis_ratio': value['min_to_maj_axis_ratio'],
                             'int_to_maj_axis_ratio': value['int_to_maj_axis_ratio'],
                             'euler_angle1': value['euler_angle1'].to('deg'),
                             'euler_angle2': value['euler_angle2'].to('deg'),
                             'euler_angle3': value['euler_angle3'].to('deg')}
        
        
    #========== Cluster boundary
    @R_truncation.setter
    def R_truncation(self, value):
        # Check type
        try:
            test = value.to('kpc')
        except:
            raise TypeError("The radius R_truncation should be a quantity homogeneous to kpc.")

        # check value
        #if value <= self._R500 :
        #    raise ValueError("Radius R_truncation should be larger than R500 for internal consistency.")

        # Set parameters
        self._R_truncation = value
        self._theta_truncation = ((self._R_truncation / self._D_ang).to('') * u.rad).to('deg')
        
        # Information
        if not self._silent: print("Setting R_truncation value")
        if not self._silent: print("Setting: theta_truncation ; Fixing: D_ang")
        
    @theta_truncation.setter
    def theta_truncation(self, value):
        # Check type
        try:
            test = value.to('deg')
        except:
            raise TypeError("The angle theta_truncation should be a quantity homogeneous to deg.")
        
        # check value
        if value <= self._theta500 :
            raise ValueError("Angle theta_truncation should be larger than theta500 for internal consistency.")

        # Set parameters
        self._theta_truncation = value
        self._R_truncation = value.to_value('rad') * self._D_ang
        
        # Information
        if not self._silent: print("Setting theta_truncation value")
        if not self._silent: print("Setting: R_truncation ; Fixing: D_ang")

    @Rmin.setter
    def Rmin(self, value):
        # Check type
        try:
            test = value.to('kpc')
        except:
            raise TypeError("The radius Rmin should be a quantity homogeneous to kpc.")

        # Check value
        if value.to_value('kpc') <= 0:
            raise TypeError("Rmin cannot be 0 (or less than 0) because integrations are in log space.")

        if value.to_value('kpc') < 1e-2:
            if not self._silent: 
                print("WARNING: the requested value of Rmin is very small. Rmin~kpc is expected")

        # Set parameters
        self._Rmin = value
        
        # Information
        if not self._silent: print("Setting Rmin value")

    #========== ICM composition
    @helium_mass_fraction.setter
    def helium_mass_fraction(self, value):
        # Check type
        if type(value) != float and type(value) != int and type(value) != np.float64:
            raise TypeError("The helium mass fraction should be a float or int")

        # Check value
        if value > 1.0 or value < 0.0:
            raise ValueError("The helium mass fraction should be between 0 and 1")

        # Set parameters
        self._helium_mass_fraction = value
        
        # Information
        if not self._silent: print("Setting helium mass fraction value")
        
    @metallicity_sol.setter
    def metallicity_sol(self, value):
        # Check type
        if type(value) != float and type(value) != int and type(value) != np.float64:
            raise TypeError("The metallicity should be a float")
        
        # Check value
        if value < 0.0:
            raise ValueError("The metallicity should be >= 0")

        # Set parameters
        self._metallicity_sol = value
        
        # Information
        if not self._silent: print("Setting metallicity value")

    @abundance.setter
    def abundance(self, value):
        # Check type
        if type(value) != float and type(value) != int and type(value) != np.float64:
            raise TypeError("The abundance should be a float")

        # Check value
        if value < 0.0:
            raise ValueError("The abundance should be >= 0")

        # Set parameters
        self._abundance = value
        
        # Information
        if not self._silent: print("Setting abundance value")
        
    #========== ICM physics
    @model_pressure_profile.setter
    def model_pressure_profile(self, value):
        # check type
        if type(value) != dict :
            raise TypeError("The model of pressure profile should be a dictionary with relevant keys")
        
        # Check the input parameters and use it
        Ppar = self._validate_model_profile_parameters(value, 'keV cm-3')
        self._model_pressure_profile = Ppar
        
        # Information
        if not self._silent: print("Setting model_pressure_profile value")
        if not self._silent: print("Fixing: R500 if involved")

    @model_pressure_fluctuation.setter
    def model_pressure_fluctuation(self, value):
        # check type
        if type(value) != dict :
            raise TypeError("The model of pressure fluctuation should be a dictionary with relevant keys")
        
        # Check the input parameters and use it
        Ppar = self._validate_model_fluctuation_parameters(value)
        self._model_pressure_fluctuation = Ppar
        
        # Information
        if not self._silent: print("Setting model_pressure_fluctuation value")
        if not self._silent: print("Fixing: R500 if involved")

        
    #========== Sampling
    @map_center.setter
    def map_center(self, value):
        err_msg = ("The coordinates can be a coord object, "
                   "or a {'RA','Dec'} or {'Glon', 'Glat'} dictionary.")
        
        # Case value is a SkyCoord object
        if type(value) == astropy.coordinates.sky_coordinate.SkyCoord:
            self._map_center = value
    
        # Case value is standard coordinates
        elif type(value) == dict:
            
            # It is not possible to have both RA-Dec and Glat-Glon, or just RA and not Dec, etc
            cond1 = 'RA'  in list(value.keys()) and 'Glat' in list(value.keys())
            cond2 = 'RA'  in list(value.keys()) and 'Glon' in list(value.keys())
            cond3 = 'Dec' in list(value.keys()) and 'Glat' in list(value.keys())
            cond4 = 'Dec' in list(value.keys()) and 'Glon' in list(value.keys())
            if cond1 or cond2 or cond3 or cond4:
                raise ValueError(err_msg)
            
            # Case where RA-Dec is used
            if 'RA' in list(value.keys()) and 'Dec' in list(value.keys()):
                self._map_center = SkyCoord(value['RA'], value['Dec'], frame="icrs")
    
            # Case where Glon-Glat is used
            elif 'Glon' in list(value.keys()) and 'Glat' in list(value.keys()):
                self._map_center = SkyCoord(value['Glon'], value['Glat'], frame="galactic")
    
            # Otherwise, not appropriate value
            else:
                raise TypeError(err_msg)
    
        # Case value is not accepted
        else:
            raise TypeError(err_msg)
    
        # Header to None
        self._map_header = None
    
        # Information
        if not self._silent: print("Setting the map coordinates")
        if not self._silent: print("Setting: map_header to None, as map properties are now set by hand")

    @map_reso.setter
    def map_reso(self, value):
        # check type
        try:
            test = value.to('deg')
        except:
            raise TypeError("The map resolution should be a quantity homogeneous to deg.")
    
        # check value
        if type(value.value) != float and type(value.value) != int and type(value.value) != np.float64:        
            raise TypeError("The map resolution should be a scalar, e.i. reso_x = reso_y.")
    
        # Set parameters
        self._map_reso = value
        self._map_header = None
        
        # Information
        if not self._silent: print("Setting the map resolution value")
        if not self._silent: print("Setting: map_header to None, as map properties are now set by hand")

    @map_fov.setter
    def map_fov(self, value):
        # check type
        try:
            test = value.to('deg')
        except:
            raise TypeError("The map field of view should be a quantity homogeneous to deg.")
    
        # Set parameters for single value application
        if type(value.value) == float or type(value.value) == int or type(value.value) == np.float64 :        
            self._map_fov = [value.to_value('deg'), value.to_value('deg')] * u.deg
    
        # Set parameters for single value application
        elif type(value.value) == np.ndarray:
            # check the dimension
            if len(value) == 2:
                self._map_fov = value
            else:
                raise TypeError("The map field of view is either a scalar, or a 2d list quantity.")
    
        # No other options
        else:
            raise TypeError("The map field of view is either a scalar, or a 2d list quantity.")
    
        # Set extra parameters
        self._map_header = None
    
        # Information
        if not self._silent: print("Setting the map field of view")
        if not self._silent: print("Setting: map_header to None, as map properties are now set by hand")

    @map_header.setter
    def map_header(self, value):
        # Check the header by reading it with WCS
        try:
            w = WCS(value)
            data_tpl = np.zeros((value['NAXIS2'], value['NAXIS1']))            
            header = w.to_header()
            hdu = fits.PrimaryHDU(header=header, data=data_tpl)
            header = hdu.header
        except:
            raise TypeError("Issue detected with the header, may not contain NAXIS1,2.")
        
        # set the value
        self._map_header = header
        self._map_center = None
        self._map_reso   = None
        self._map_fov    = None
    
        # Information
        if not self._silent: print("Setting the map header")
        if not self._silent: print("Setting: map_center, map_reso, map_fov to None, as the header will be used")

    @los_reso.setter
    def los_reso(self, value):
        # check type
        try:
            test = value.to('kpc')
        except:
            raise TypeError("The l.o.s. resolution should be a quantity homogeneous to kpc.")
    
        # check value
        if type(value.value) != float and type(value.value) != int and type(value.value) != np.float64:        
            raise TypeError("The l.o.s. resolution should be a scalar.")
    
        # Set parameters
        self._los_reso = value
        
        # Information
        if not self._silent: print("Setting the l.o.s. resolution value")

    @los_size.setter
    def los_size(self, value):
        # check type
        try:
            test = value.to('kpc')
        except:
            raise TypeError("The l.o.s. size should be a quantity homogeneous to kpc.")
    
        # check value
        if type(value.value) != float and type(value.value) != int and type(value.value) != np.float64:        
            raise TypeError("The l.o.s. size should be a scalar.")
    
        # Set parameters
        self._los_size = value
        
        # Information
        if not self._silent: print("Setting the l.o.s size value")




"""
This file deals with the radial model fitting

"""

import numpy as np
import copy
import astropy.units as u
from astropy.coordinates import SkyCoord
from pathos.multiprocessing import ProcessPool
from multiprocessing import Pool, cpu_count
import emcee
from scipy.optimize import curve_fit
from minot.ClusterTools.map_tools import radial_profile_sb

from pitszi import utils_fitting
from pitszi import utils_plot


#==================================================
# InferenceFitting class
#==================================================

class InferenceRadialFitting(object):
    """ InferenceRadialFitting class
    This class serves as a parser to the main Inference class, to
    include the subclass InferenceRadialFitting in this other file.

    Attributes
    ----------
    The attributes are the same as the Inference class, see inference_radial_main.py

    Methods
    ----------
    - get_mcmc_chains_outputs_results: utility tools to get mcmc chain results
    - get_curvefit_outputs_results: utility tools to get curvefit results
    - defpar_profile: define the profile parameters
    - prior_profile: define the prior given the parameter
    - setpar_profile: set the model with the parameters
    - lnlike_profile: likelihood function definition
    - run_mcmc_profile: run the mcmc fit
    - run_mcmc_profile_results: produce the results given the mcmc fit
    - run_curvefit_profile: run the curvefit
    - run_curvefit_profile_results: produce the results given the curvefit

    """

    #==================================================
    # Compute results for a given MCMC sampling
    #==================================================
    
    def get_mcmc_chains_outputs_results(self,
                                       parlist,
                                       sampler,
                                       conf=68.0,
                                       truth=None,
                                       extraname=''):
        """
        This function can be used to produce automated plots/files
        that give the results of the MCMC samping
        
        Parameters
        ----------
        - parlist (list): the list of parameter names
        - sampler (emcee object): the sampler
        - conf (float): confidence limit in % used in results
        - truth (list): the list of expected parameter value for the fit
        - extraname (string): extra name to add after MCMC in file name

        Outputs
        ----------
        Plots are produced

        """

        Nchains, Nsample, Nparam = sampler.chain.shape
        
        #---------- Check the truth
        if truth is not None:
            if len(truth) != Nparam:
                raise ValueError("The 'truth' keyword should match the number of parameters")

        #---------- Remove the burnin
        if self.mcmc_burnin <= Nsample:
            par_chains = sampler.chain[:,self.mcmc_burnin:,:]
            lnl_chains = sampler.lnprobability[:,self.mcmc_burnin:]
        else:
            par_chains = sampler.chain
            lnl_chains = sampler.lnprobability
            if not self.silent:
                print('The burnin could not be remove because it is larger than the chain size')
            
        #---------- Compute statistics for the chains
        utils_fitting.chains_statistics(par_chains, lnl_chains,
                                        parname=parlist,
                                        conf=conf,
                                        outfile=self.output_dir+'/MCMC'+extraname+'_chain_statistics.txt')

        #---------- Produce 1D plots of the chains
        utils_plot.chains_1Dplots(par_chains, parlist, self.output_dir+'/MCMC'+extraname+'_chain_1d_plot.pdf')
        
        #---------- Produce 1D histogram of the chains
        namefiles = [self.output_dir+'/MCMC'+extraname+'_chain_hist_'+i+'.pdf' for i in parlist]
        utils_plot.chains_1Dhist(par_chains, parlist, namefiles,
                                 conf=conf, truth=truth)

        #---------- Produce 2D (corner) plots of the chains
        utils_plot.chains_2Dplots_corner(par_chains,
                                         parlist,
                                         self.output_dir+'/MCMC'+extraname+'_chain_2d_plot_corner.pdf',
                                         truth=truth)

        #utils_plot.chuains_2Dplots_sns(par_chains,
        #                              parlist,
        #                              self.output_dir+'/MCMC'+extraname+'_chain_2d_plot_sns.pdf',
        #                              truth=truth)
    

    #==================================================
    # Compute results for a given curvefit
    #==================================================
    
    def get_curvefit_outputs_results(self,
                                     parlist,
                                     parinfo,
                                     popt, pcov,
                                     conf=68.0,
                                     truth=None,
                                     extraname='',
                                     Nsample=10000):
        """
        This function can be used to produce automated plots/files
        that give the results of the curvefit fit
        
        Parameters
        ----------
        - parlist (list): the list of parameter names
        - popt (1d array): best fit parameter values
        - pcov (2d array): posterior covariance matrix
        - conf (float): confidence limit in % used in results
        - truth (list): the list of expected parameter value for the fit
        - extraname (string): extra name to add after MCMC in file name

        Outputs
        ----------
        Plots are produced

        """

        Nparam = len(parlist)
        Nchains = 1
        
        #---------- Check the truth
        if truth is not None:
            if len(truth) != Nparam:
                raise ValueError("The 'truth' keyword should match the number of parameters")

        #---------- Mimic MCMC chains with multivariate Gaussian
        par_chains = np.zeros((Nsample, Nparam))
        isamp = 0
        while isamp < Nsample:
            param = np.random.multivariate_normal(popt, pcov)
            cond = np.isfinite(self.prior_profile(param, parinfo)) # make sure params are within limits
            if cond:
                par_chains[isamp,:] = param
                isamp += 1
        
        lnl_chains = np.zeros(Nsample)
        for i in range(Nsample):
            lnl_chains[i] = -0.5 * np.matmul((par_chains[i,:]-popt), np.matmul(pcov, (par_chains[i,:]-popt)))
        par_chains = par_chains[np.newaxis]
        lnl_chains = lnl_chains[np.newaxis]
            
        #---------- Compute statistics for the parameters
        utils_fitting.chains_statistics(par_chains, lnl_chains,
                                        parname=parlist,
                                        conf=conf,
                                        outfile=self.output_dir+'/CurveFit'+extraname+'_statistics.txt')
            
        #---------- Produce 1D plots of the chains
        utils_plot.chains_1Dplots(par_chains, parlist, self.output_dir+'/CurveFit'+extraname+'_1d_plot.pdf')
        
        #---------- Produce 1D histogram of the chains
        namefiles = [self.output_dir+'/CurveFit'+extraname+'_hist_'+i+'.pdf' for i in parlist]
        utils_plot.chains_1Dhist(par_chains, parlist, namefiles,
                                 conf=conf, truth=truth)

        #---------- Produce 2D (corner) plots of the chains
        utils_plot.chains_2Dplots_corner(par_chains,
                                         parlist,
                                         self.output_dir+'/CurveFit'+extraname+'_2d_plot_corner.pdf',
                                         truth=truth)

        #utils_plot.chains_2Dplots_sns(par_chains,
        #                              parlist,
        #                              self.output_dir+'/CurveFit'+extraname+'_2d_plot_sns.pdf',
        #                              truth=truth)
    
        
    #==================================================
    # Profile parameter definition
    #==================================================
        
    def defpar_profile(self, parinfo):
        """
        This function helps defining the parameter list, initial values and
        ranges for the profile fitting.
        We distinguish 4 kind of parameters:
        - profile parameters or M500
        - offset RA and Dec 
        - ellipticity min_to_maj_axis_ratio and angle
        - nuisance zero level ZL
            
        Parameters
        ----------
         - parinfo (dict): parameter information
    
        Outputs
        ----------
        - par_list (list): list of parameter names
        - par0_value (list): list of mean initial parameter value
        - par0_err (list): list of mean initial parameter uncertainty
        - par_min (list): list of minimal initial parameter value
        - par_max (list): list of maximal initial parameter value
        
        """
        
        parkeys = list(parinfo.keys())
        
        #========== Init the parameters
        par_list = []
        par0_value = []
        par0_err = []
        par_min = []
        par_max = []
        
        #========== Check the profile parameters
        # remove unwanted keys
        parinfo_prof = copy.deepcopy(parinfo)
        if 'RA'                    in parkeys: parinfo_prof.pop('RA')
        if 'Dec'                   in parkeys: parinfo_prof.pop('Dec')
        if 'min_to_maj_axis_ratio' in parkeys: parinfo_prof.pop('min_to_maj_axis_ratio')
        if 'angle'                 in parkeys: parinfo_prof.pop('angle')
        if 'ZL'                    in parkeys: parinfo_prof.pop('ZL')

        # Check that if M500 is a parameter, it is the only one
        Npar_prof = len(list(parinfo_prof.keys()))
        if 'M500' in list(parinfo_prof.keys()) and Npar_prof > 1:
            raise ValueError('M500 is a fit parameter, thus not other profile parameter is accepted.')    
            
        # Loop over the keys
        for ipar in range(Npar_prof):
            
            parkey = list(parinfo_prof.keys())[ipar]

            #----- Check that parameters are working
            try:
                if 'M500' == parkey:
                    bid = self.model.M500
                else:
                    bid = self.model.model_pressure_profile[parkey]
            except:
                raise ValueError('The parameter '+parkey+' is not in self.model.model_pressure_profile')    

            if 'M500' == parkey:
                try:
                    bid = parinfo_prof[parkey]['P_ref']
                except:
                    raise ValueError('M500 dict should contain "P_ref":{UPP, A10UPP, etc} passed to set_pressure_profile_universal_param')    

            if 'guess' not in parinfo_prof[parkey]:
                raise ValueError('The guess key is mandatory for starting the chains, as "guess":[guess_value, guess_uncertainty] ')    
            if 'unit' not in parinfo_prof[parkey]:
                raise ValueError('The unit key is mandatory. Use "unit":None if unitless')
    
            #----- Update the name list
            par_list.append(parkey)
    
            #----- Update the guess values
            par0_value.append(parinfo_prof[parkey]['guess'][0])
            par0_err.append(parinfo_prof[parkey]['guess'][1])
    
            #----- Update the limit values
            if 'limit' in parinfo_prof[parkey]:
                par_min.append(parinfo_prof[parkey]['limit'][0])
                par_max.append(parinfo_prof[parkey]['limit'][1])
            else:
                par_min.append(-np.inf)
                par_max.append(+np.inf)

        #========== Center parameters
        # Collect RA, Dec keys
        list_ctr_allowed = ['RA', 'Dec']
        parinfo_center = {key: parinfo[key] for key in list_ctr_allowed if key in parinfo.keys()}
        Npar_ctr = len(list(parinfo_center.keys()))
            
        # Loop over the keys
        for ipar in range(Npar_ctr):
            
            parkey = list(parinfo_center.keys())[ipar]
            
            #----- Check that the param is ok wrt the model
            if 'guess' not in parinfo_center[parkey]:
                raise ValueError('The guess key is mandatory for starting the chains, as "guess":[guess_value, guess_uncertainty] ')    
            if 'unit' not in parinfo_center[parkey]:
                raise ValueError('The unit key is mandatory. Use "unit":None if unitless')
            
            #----- Update the name list
            par_list.append(parkey)
            
            #----- Update the guess values
            par0_value.append(parinfo_center[parkey]['guess'][0])
            par0_err.append(parinfo_center[parkey]['guess'][1])
            
            #----- Update the limit values
            if 'limit' in parinfo_center[parkey]:
                par_min.append(parinfo_center[parkey]['limit'][0])
                par_max.append(parinfo_center[parkey]['limit'][1])
            else:
                par_min.append(-np.inf)
                par_max.append(+np.inf)
    
        #========== Ellipticity parameters
        list_ell_allowed = ['min_to_maj_axis_ratio', 'angle']
        parinfo_ellipticity = {key: parinfo[key] for key in list_ell_allowed if key in parinfo.keys()}
        Npar_ell = len(list(parinfo_ellipticity.keys()))

        #----- Check that both parameter are fitted
        if Npar_ell == 1:
            raise ValueError('Ellipticity parameter (min_to_maj_axis_ratio and angle) cannot be fitted separately')

        #----- Loop
        for ipar in range(Npar_ell):
    
            parkey = list(parinfo_ellipticity.keys())[ipar]
    
            #----- Check that the param is ok wrt the model
            if 'guess' not in parinfo_ellipticity[parkey]:
                raise ValueError('The guess key is mandatory for starting the chains, as "guess":[guess_value, guess_uncertainty]')    
            if 'unit' not in parinfo_ellipticity[parkey]:
                raise ValueError('The unit key is mandatory. Use "unit":None if unitless')
    
            #----- Update the name list
            par_list.append(parkey)
    
            #----- Update the guess values
            par0_value.append(parinfo_ellipticity[parkey]['guess'][0])
            par0_err.append(parinfo_ellipticity[parkey]['guess'][1])
    
            #----- Update the limit values
            if 'limit' in parinfo_ellipticity[parkey]:
                par_min.append(parinfo_ellipticity[parkey]['limit'][0])
                par_max.append(parinfo_ellipticity[parkey]['limit'][1])
            else:
                if parkey == 'min_to_maj_axis_ratio':
                    par_min.append(0)
                    par_max.append(1)
                if parkey == 'angle':
                    par_min.append((-90*u.deg).to_value(parinfo_ellipticity[parkey]['unit']))
                    par_max.append((+90*u.deg).to_value(parinfo_ellipticity[parkey]['unit']))
    
        #========== Map offset parameter
        list_zl_allowed = ['ZL']
        parinfo_ZL = {key: parinfo[key] for key in list_zl_allowed if key in parinfo.keys()}
        Npar_zl = len(list(parinfo_ZL.keys()))
    
        for ipar in range(Npar_zl):
    
            parkey = list(parinfo_ZL.keys())[ipar]
    
            #----- Check that the param is ok wrt the model
            if 'guess' not in parinfo_ZL[parkey]:
                raise ValueError('The guess key is mandatory for starting the chains, as "guess":[guess_value, guess_uncertainty] ')    
            if 'unit' not in parinfo_ZL[parkey]:
                raise ValueError('The unit key is mandatory. Use "unit":None if unitless')
    
            #----- Update the name list
            par_list.append(parkey)
    
            #----- Update the guess values
            par0_value.append(parinfo_ZL[parkey]['guess'][0])
            par0_err.append(parinfo_ZL[parkey]['guess'][1])
    
            #----- Update the limit values
            if 'limit' in parinfo_ZL[parkey]:
                par_min.append(parinfo_ZL[parkey]['limit'][0])
                par_max.append(parinfo_ZL[parkey]['limit'][1])
            else:
                par_min.append(-np.inf)
                par_max.append(+np.inf)
            
        #========== Make it an np array
        par_list   = np.array(par_list)
        par0_value = np.array(par0_value)
        par0_err   = np.array(par0_err)
        par_min    = np.array(par_min)
        par_max    = np.array(par_max)
        
        return par_list, par0_value, par0_err, par_min, par_max

        
    #==================================================
    # Profile prior definition
    #==================================================
        
    def prior_profile(self, param, parinfo):
        """
        Compute the prior given the input parameter and some information
        about the parameters given by the user
        We distinguish 4 kind of parameters:
        - profile parameters
        - offset RA and Dec 
        - ellipticity min_to_maj_axis_ratio and angle
        - nuisance zero level ZL
        
         Parameters
         ----------
         - param (list): the parameter to apply to the model
         - parinfo (dict): profile parameter information
     
         Outputs
         ----------
         - prior (float): value of the prior

        """

        prior = 0
        idx_par = 0
        parkeys = list(parinfo.keys())

        #========== Check the profile parameters
        # remove unwanted keys
        parinfo_prof = copy.deepcopy(parinfo)
        if 'RA'                    in parkeys: parinfo_prof.pop('RA')
        if 'Dec'                   in parkeys: parinfo_prof.pop('Dec')
        if 'min_to_maj_axis_ratio' in parkeys: parinfo_prof.pop('min_to_maj_axis_ratio')
        if 'angle'                 in parkeys: parinfo_prof.pop('angle')
        if 'ZL'                    in parkeys: parinfo_prof.pop('ZL')
        
        # Loop over the keys
        parkeys_prof = list(parinfo_prof.keys())

        for ipar in range(len(parkeys_prof)):
            parkey = parkeys_prof[ipar]
        
            # Flat prior
            if 'limit' in parinfo_prof[parkey]:
                if param[idx_par] < parinfo_prof[parkey]['limit'][0]:
                    return -np.inf                
                if param[idx_par] > parinfo_prof[parkey]['limit'][1]:
                    return -np.inf
                
            # Gaussian prior
            if 'prior' in parinfo_prof[parkey]:
                expected = parinfo_prof[parkey]['prior'][0]
                sigma = parinfo_prof[parkey]['prior'][1]
                prior += -0.5*(param[idx_par] - expected)**2 / sigma**2
    
            # Increase param index
            idx_par += 1

        #========== Other parameters
        list_allowed  = ['RA', 'Dec', 'min_to_maj_axis_ratio', 'angle', 'ZL']
        parinfo_other = {key: parinfo[key] for key in list_allowed if key in parinfo.keys()}        
        parkeys_other = list(parinfo_other.keys())
        Npar_other    = len(parkeys_other)

        for ipar in range(Npar_other):
            parkey = parkeys_other[ipar]
            
            # Flat prior
            if 'limit' in parinfo_other[parkey]:
                if param[idx_par] < parinfo_other[parkey]['limit'][0]:
                    return -np.inf                
                if param[idx_par] > parinfo_other[parkey]['limit'][1]:
                    return -np.inf
                
            # Gaussian prior
            if 'prior' in parinfo_other[parkey]:
                expected = parinfo_other[parkey]['prior'][0]
                sigma = parinfo_other[parkey]['prior'][1]
                prior += -0.5*(param[idx_par] - expected)**2 / sigma**2
     
            # Increase param index
            idx_par += 1

        #========== Check on param numbers
        if idx_par != len(param):
            raise ValueError('Problem with the prior parameters')
        
        return prior
    
    
    #==================================================
    # Profile parameter settings
    #==================================================
        
    def setpar_profile(self, param, parinfo):
        """
        Set the model to the given profile parameters.
        We distinguish 4 kind of parameters:
        - profile parameters
        - offset RA and Dec 
        - ellipticity min_to_maj_axis_ratio and angle
        - nuisance zero level ZL        

        Parameters
        ----------
        - param (list): the parameter to apply to the model
        - parinfo (dict): see mcmc_profile

        Outputs
        ----------
        The model is updated with parameter settings

        """
        
        idx_par = 0
        parkeys = list(parinfo.keys())

        #========== Check the profile parameters
        # remove unwanted keys
        parinfo_prof = copy.deepcopy(parinfo)
        if 'RA'                    in parkeys: parinfo_prof.pop('RA')
        if 'Dec'                   in parkeys: parinfo_prof.pop('Dec')
        if 'min_to_maj_axis_ratio' in parkeys: parinfo_prof.pop('min_to_maj_axis_ratio')
        if 'angle'                 in parkeys: parinfo_prof.pop('angle')
        if 'ZL'                    in parkeys: parinfo_prof.pop('ZL')
        parkeys_prof = list(parinfo_prof.keys())
        
        # Loop Profile
        for ipar in range(len(parkeys_prof)):
            parkey = parkeys_prof[ipar]
            if parinfo_prof[parkey]['unit'] is not None:
                unit = parinfo_prof[parkey]['unit']
            else:
                unit = 1

            if parkey == 'M500':
                self.model.M500 = param[idx_par] * unit
                self.model.set_pressure_profile_universal_param(parinfo['M500']['P_ref'])
            else:
                self.model.model_pressure_profile[parkey] = param[idx_par] * unit
            idx_par += 1

        #========== Center
        if 'RA' in parkeys:
            RA = param[idx_par]
            unit1 = parinfo['RA']['unit']
            idx_par += 1
        else:
            RA = self.model.coord.icrs.ra.to_value('deg')
            unit1 = u.deg
            
        if 'Dec' in parkeys:
            Dec = param[idx_par]
            unit2 = parinfo['Dec']['unit']
            idx_par += 1
        else:
            Dec = self.model.coord.icrs.dec.to_value('deg')
            unit2 = u.deg

        self.model.coord = SkyCoord(RA*unit1, Dec*unit2, frame='icrs')
                
        #========== Ellipticity
        if 'min_to_maj_axis_ratio' in parkeys and 'angle' in parkeys:
            axis_ratio = param[idx_par]
            idx_par += 1
            
            angle = param[idx_par]
            angle_unit = parinfo['angle']['unit']
            idx_par += 1
            
            self.model.triaxiality = {'min_to_maj_axis_ratio':axis_ratio,
                                      'int_to_maj_axis_ratio':axis_ratio,
                                      'euler_angle1':0*u.deg,
                                      'euler_angle2':90*u.deg,
                                      'euler_angle3':angle*angle_unit}
        
        #========== Map zero level
        if 'ZL' in parkeys:
            self.nuisance_ZL = param[idx_par]
            idx_par += 1

        #========== Final check on parameter count
        if len(param) != idx_par:
            print('issue with the number of parameters')
            import pdb
            pdb.set_trace()

        
    #==================================================
    # lnL function for the profile fit
    #==================================================
    
    def lnlike_profile(self, param, parinfo):
        """
        This is the likelihood function used for the fit of the profile.
            
        Parameters
        ----------
        - param (np array): the value of the test parameters
        - parinfo (dict): see parinfo in mcmc_profile

        Outputs
        ----------
        - lnL (float): the value of the log likelihood
    
        """
        
        #========== Deal with flat and Gaussian priors
        prior = self.prior_profile(param, parinfo)
        if not np.isfinite(prior): return -np.inf
        
        #========== Change model parameters
        self.setpar_profile(param, parinfo)
        
        #========== Get the model
        model_img = self.get_radial_model()
        
        #========== Compute the likelihood
        if self.method_use_covmat:
            flat_resid = (self.data.mask * (model_img - self.data.image)).flatten()
            lnL = -0.5*np.matmul(flat_resid, np.matmul(self._ymap_invcov, flat_resid))
        else:
            lnL = -0.5*np.nansum(self.data.mask**2 * (model_img-self.data.image)**2 / self.data.noise_rms**2)
            
        lnL += prior
        
        #========== Check and return
        if np.isnan(lnL):
            return -np.inf
        else:
            return lnL        

        
    #==================================================
    # Compute the smooth model via forward fitting
    #==================================================
    
    def run_mcmc_profile(self, parinfo,
                         filename_sampler=None,
                         show_fit_result=False,
                         set_bestfit=False):
        """
        This function fits the data given the current model and the parameter information
        given by the user, using MCMC.
        
        Parameters
        ----------
        - parinfo (dictionary): the model parameters associated with 
        self.model.model_pressure_profile to be fit as, e.g., 
        parinfo =      {'P_0':                   # --> Parameter key (mandatory)
                        {'guess':[0.01, 0.001],  # --> initial guess: center, uncertainty (mandatory)
                         'unit': u.keV*u.cm**-3, # --> unit (mandatory, None if unitless)
                         'limit':[0, np.inf],    # --> Allowed range, i.e. flat prior (optional)
                         'prior':[0.01, 0.001],  # --> Gaussian prior: mean, sigma (optional)
                        },
                        'r_p':
                        {'limit':[0, np.inf], 
                          'prior':[100, 1],
                          'unit': u.kpc, 
                        },
                       }    
        Other accepted parameters: 'RA' and 'Dec', 'min_to_maj_axis_ratio' and 'angle', 'ZL'

        Instead of the profile parameters, the mass M500 is also accepted but the M500 disctionary 
        should also contain P_ref as the reference profile model, e.g. A10UPP, etc, to be passed to
        model.set_pressure_profile_universal_param : 'M500':{'guess':..., 'P_ref': 'A10UPP'}

        - filename_sampler (str): the file name of the sampler to use.
        - show_fit_result (bool): show the best fit model and residual.
        - set_bestfit (bool): set the best fit to the model

        Outputs
        ----------
        - parlist (list): the list of the fit parameters
        - sampler (emcee object): the sampler associated with model parameters of
        the smooth component

        """

        #========== Copy the input model
        input_model = copy.deepcopy(self.model)
        
        #========== Check if the MCMC sampler was already recorded
        if filename_sampler is None:
            sampler_file = self.output_dir+'/pitszi_MCMC_Profile_sampler.h5'
        else:
            sampler_file = filename_sampler
            
        sampler_exist = utils_fitting.check_sampler_exist(sampler_file, silent=False)
        
        #========== Defines the fit parameters
        # Fit parameter list and information
        par_list, par0_value, par0_err, par_min, par_max = self.defpar_profile(parinfo)
        ndim = len(par0_value)
        
        # Starting points of the chains
        pos = utils_fitting.emcee_starting_point(par0_value, par0_err, par_min, par_max, self.mcmc_nwalkers)
        if sampler_exist and (not self.mcmc_reset): pos = None

        # Parallel mode
        if self.method_parallel:
            mypool = ProcessPool()
        else:
            mypool = None

        # Get the inverse covariance matrix if needed
        if self.method_use_covmat:
            if not self.silent:
                print('Using the ymap covariance matrix may failmay be unstable. Starting to compute C^-1')
            self._ymap_invcov = np.linalg.inv(self.data.noise_covmat)
                
        # Info
        if not self.silent:
            print('----- Fit parameters information -----')
            print('      - Fitted parameters:            ')
            print(par_list)
            print('      - Starting point mean:          ')
            print(par0_value)
            print('      - Starting point dispersion :   ')
            print(par0_err)
            print('      - Minimal starting point:       ')
            print(par_min)
            print('      - Maximal starting point:       ')
            print(par_max)
            print('      - Number of dimensions:         ')
            print(ndim)
            print('      - Parallel mode:                ')
            print(self.method_parallel)
            print('      - Use covariance matrix?        ')
            print(self.method_use_covmat)
            print('-----')

        #========== Define the MCMC setup
        backend = utils_fitting.define_emcee_backend(sampler_file, sampler_exist,
                                                     self.mcmc_reset, self.mcmc_nwalkers, ndim, silent=False)
        moves = emcee.moves.StretchMove(a=2.0)
        
        sampler = emcee.EnsembleSampler(self.mcmc_nwalkers, ndim,
                                        self.lnlike_profile, 
                                        args=[parinfo], 
                                        pool=mypool,
                                        moves=moves,
                                        backend=backend)

        #========== Run the MCMC
        if not self.silent: print('----- MCMC sampling -----')
        if self.mcmc_run:
            if not self.silent: print('      - Runing '+str(self.mcmc_nsteps)+' MCMC steps')
            res = sampler.run_mcmc(pos, self.mcmc_nsteps, progress=True, store=True)
        else:
            if not self.silent: print('      - Not running, but restoring the existing sampler')

        #========== Show results
        if show_fit_result:
            self.get_mcmc_chains_outputs_results(par_list, sampler, extraname='_Profile')
            self.run_mcmc_profile_results(sampler, parinfo)

        #========== Compute the best-fit model and set it
        if set_bestfit:
            best_par = utils_fitting.get_emcee_bestfit_param(sampler, self.mcmc_burnin)
            self.setpar_profile(best_par, parinfo)
        else:
            self.model = input_model
            
        return par_list, sampler

    
    #==================================================
    # Show the fit results related to profile
    #==================================================
    
    def run_mcmc_profile_results(self, sampler, parinfo,
                                 visu_smooth=10*u.arcsec,
                                 binsize_prof=5*u.arcsec,
                                 true_pressure_profile=None,
                                 true_compton_profile=None):
        """
        This is function is used to show the results of the MCMC
        regarding the radial profile
            
        Parameters
        ----------
        - sampler (emcee object): the sampler obtained from mcmc_profile
        - parinfo (dict): same as mcmc_profile
        - visu_smooth (quantity): The extra smoothing FWHM for vidualization. Homogeneous to arcsec.
        - binsize_prof (quantity): The binsize for the y profile. Homogeneous to arcsec
        - true_pressure_profile (dict): pass a dictionary containing the profile to compare with
        in the form {'r':array in kpc, 'p':array in keV cm-3}
        - true_compton_profile (dict): pass a dictionary containing the profile to compare with
        in the form {'r':array in arcmin, 'y':array in [y]}
        
        Outputs
        ----------
        plots are produced

        """

        #========== Get noise MC 
        noise_mc = self.data.noise_mc
        if noise_mc.shape[0] > self.mcmc_Nresamp:
            Nmc = self.mcmc_Nresamp
            noise_mc = noise_mc[0:Nmc]
        else:
            Nmc = noise_mc.shape[0]
            if self.silent == False: print('WARNING: the number of noise MC is lower than requested')

        #========== rms for profile
        rms_prof = np.std(noise_mc, axis=0)
        mymask = self.data.mask**2
        mymask[self.data.mask == 0] = np.nan
        rms_prof = rms_prof/mymask
        rms_prof[~np.isfinite(rms_prof)] = np.nan
        
        #========== Get the best-fit
        best_par = utils_fitting.get_emcee_bestfit_param(sampler, self.mcmc_burnin)
        self.setpar_profile(best_par, parinfo)
        best_model = copy.deepcopy(self.model)

        #========== Get the profile for the data, centered on best-fit center if fitted
        data_yprof_tmp = radial_profile_sb(self.data.image, 
                                           (best_model.coord.icrs.ra.to_value('deg'),
                                            best_model.coord.icrs.dec.to_value('deg')), 
                                           stddev=rms_prof, header=self.data.header, 
                                           binsize=binsize_prof.to_value('deg'))
        r2d = data_yprof_tmp[0]*60
        data_yprof = data_yprof_tmp[1]

        mc_y_data = np.zeros((Nmc, len(r2d)))
        for i in range(Nmc):
            mc_y_data[i,:] = radial_profile_sb(noise_mc[i,:,:], 
                                               (best_model.coord.icrs.ra.to_value('deg'),
                                                best_model.coord.icrs.dec.to_value('deg')), 
                                               stddev=rms_prof, header=self.data.header, 
                                               binsize=binsize_prof.to_value('deg'))[1]
        data_yprof_err = np.std(mc_y_data, axis=0)

        #========== Compute best fit observables
        self.setpar_profile(best_par, parinfo)
        best_ymap_sph = self.get_radial_model()

        best_y_profile = radial_profile_sb(best_ymap_sph, 
                                           (best_model.coord.icrs.ra.to_value('deg'),
                                            best_model.coord.icrs.dec.to_value('deg')), 
                                           stddev=best_ymap_sph*0+1, header=self.data.header, 
                                           binsize=binsize_prof.to_value('deg'))[1]

        r3d, best_pressure_profile = best_model.get_pressure_profile(np.logspace(1, 4, 100)*u.kpc)
        r3d = r3d.to_value('kpc')
        best_pressure_profile = best_pressure_profile.to_value('keV cm-3')
        
        #========== MC resampling
        MC_ymap_sph         = np.zeros((self.mcmc_Nresamp, best_ymap_sph.shape[0], best_ymap_sph.shape[1]))
        MC_y_profile        = np.zeros((self.mcmc_Nresamp, len(r2d)))
        MC_pressure_profile = np.zeros((self.mcmc_Nresamp, len(r3d)))
        
        MC_pars = utils_fitting.get_emcee_random_param(sampler, burnin=self.mcmc_burnin, Nmc=self.mcmc_Nresamp)
        for i in range(self.mcmc_Nresamp):
            # Get MC model
            self.setpar_profile(MC_pars[i,:], parinfo)
            MC_ymap_sph[i,:,:] = self.get_radial_model()
            
            # Get MC y profile
            MC_y_profile[i,:] = radial_profile_sb(MC_ymap_sph[i,:,:], 
                                                  (best_model.coord.icrs.ra.to_value('deg'),
                                                   best_model.coord.icrs.dec.to_value('deg')), 
                                                  stddev=self.data.image*0+1, header=self.data.header, 
                                                  binsize=binsize_prof.to_value('deg'))[1]
            # Get MC pressure profile
            MC_pressure_profile[i,:] = self.model.get_pressure_profile(r3d*u.kpc)[1].to_value('keV cm-3')

        #========== plots map
        utils_plot.show_fit_result_ymap(self.output_dir+'/MCMC_Profile_results_y_map.pdf',
                                        self.data.image,
                                        self.data.header,
                                        noise_mc,
                                        best_ymap_sph,
                                        mask=self.data.mask,
                                        visu_smooth=visu_smooth.to_value('arcsec'))
        
        #========== plots ymap profile
        utils_plot.show_fit_ycompton_profile(self.output_dir+'/MCMC_Profile_results_y_profile.pdf',
                                             r2d, data_yprof, data_yprof_err,
                                             best_y_profile, MC_y_profile,
                                             true_compton_profile=true_compton_profile)
        
        #========== plots pressure profile
        utils_plot.show_fit_result_pressure_profile(self.output_dir+'/MCMC_Profile_results_P_profile.pdf',
                                                    r3d, best_pressure_profile, MC_pressure_profile,
                                                    true_pressure_profile=true_pressure_profile)


    #==================================================
    # Compute the smooth model via forward curvefit
    #==================================================
    
    def run_curvefit_profile(self, parinfo,
                             show_fit_result=False,
                             set_bestfit=False):
        """
        This function fits the data given the current model and the parameter information
        given by the user, using curvefit.
        
        Parameters
        ----------
        - parinfo (dictionary): the model parameters associated with 
        self.model.model_pressure_profile to be fit as, e.g., 
        parinfo =      {'P_0':                   # --> Parameter key (mandatory)
                        {'guess':[0.01, 0.001],  # --> initial guess: center, uncertainty (mandatory)
                         'unit': u.keV*u.cm**-3, # --> unit (mandatory, None if unitless)
                         'limit':[0, np.inf],    # --> Allowed range, i.e. flat prior (optional)
                         'prior':[0.01, 0.001],  # --> Gaussian prior: mean, sigma (optional)
                        },
                        'r_p':
                        {'limit':[0, np.inf], 
                          'prior':[100, 1],
                          'unit': u.kpc, 
                        },
                       }    
        Other accepted parameters: 'RA' and 'Dec', 'min_to_maj_axis_ratio' and 'angle', 'ZL'

        Instead of the profile parameters, the mass M500 is also accepted but the M500 disctionary 
        should also contain P_ref as the reference profile model, e.g. A10UPP, etc, to be passed to
        model.set_pressure_profile_universal_param : 'M500':{'guess':..., 'P_ref': 'A10UPP'}

        - show_fit_result (bool): show the best fit model and residual.

        Outputs
        ----------
        - parlist (list): the list of the fit parameters
        - par_opt (list): the best fit parameters
        - par_cov (2d matrix): the posterior covariance matrix

        """

        #========== Copy the input model
        input_model = copy.deepcopy(self.model)
        
        #========== Defines the fit parameters
        par_list, par0_value, par0_err, par_min, par_max = self.defpar_profile(parinfo)

        #========== Define the sigma
        if self.method_use_covmat:
            sigma = self.data.noise_covmat
            if not self.silent:
                print('Using the ymap covariance matrix with curve_fit is a bit ambitious, it may fail.')
        else:
            sigma = self.data.noise_rms.flatten()

        #========== Defines the data
        ymap = (self.data.image*self.data.mask).flatten()
        
        #========== Fitting function
        def fitfunc(x, *pars):
            params = []
            for ip in range(len(pars)): params.append(pars[ip])
            self.setpar_profile(np.array(params), parinfo)
            model_img = self.get_radial_model()
            model_test = model_img * self.data.mask
            return model_test.flatten()

        #========== Run the fit
        par_opt, par_cov = curve_fit(fitfunc, 0,
                                     ymap,
                                     p0=par0_value,
                                     sigma=sigma,
                                     absolute_sigma=True,
                                     bounds=(par_min, par_max))

        #========== Show results
        if show_fit_result:
            self.get_curvefit_outputs_results(par_list, parinfo, par_opt, par_cov, extraname='_Profile')
            self.run_curvefit_profile_results(par_opt, par_cov, parinfo)
        
        #========== Compute the best-fit model and set it
        if set_bestfit:
            self.setpar_profile(par_opt, parinfo)
        else:
            self.model = input_model
        
        return par_list, par_opt, par_cov
    
    
    #==================================================
    # Show the fit results related to profile
    #==================================================
    
    def run_curvefit_profile_results(self, popt, pcov, parinfo,
                                     visu_smooth=10*u.arcsec,
                                     binsize_prof=5*u.arcsec,
                                     true_pressure_profile=None,
                                     true_compton_profile=None):
        """
        This is function is used to show the results of the MCMC
        regarding the radial profile
            
        Parameters
        ----------
        - popt (1d array): parameter best fit
        - pcov (2d array): parameter covariance matrix
        - parinfo (dict): same as mcmc_profile
        - visu_smooth (quantity): The extra smoothing FWHM for vidualization. Homogeneous to arcsec.
        - binsize_prof (quantity): The binsize for the y profile. Homogeneous to arcsec
        - true_pressure_profile (dict): pass a dictionary containing the profile to compare with
        in the form {'r':array in kpc, 'p':array in keV cm-3}
        - true_compton_profile (dict): pass a dictionary containing the profile to compare with
        in the form {'r':array in arcmin, 'y':array in [y]}
        
        Outputs
        ----------
        plots are produced

        """
        
        #========== Get noise MC 
        noise_mc = self.data.noise_mc
        if noise_mc.shape[0] > self.mcmc_Nresamp:
            Nmc = self.mcmc_Nresamp
            noise_mc = noise_mc[0:Nmc]
        else:
            Nmc = noise_mc.shape[0]
            if self.silent == False: print('WARNING: the number of noise MC is lower than requested')

        #========== rms for profile
        rms_prof = np.std(noise_mc, axis=0)
        mymask = self.data.mask**2
        mymask[self.data.mask == 0] = np.nan
        rms_prof = rms_prof/mymask
        rms_prof[~np.isfinite(rms_prof)] = np.nan
        
        #========== Get the best-fit
        self.setpar_profile(popt, parinfo)
        best_model = copy.deepcopy(self.model)

        #========== Get the profile for the data, centered on best-fit center if fitted
        data_yprof_tmp = radial_profile_sb(self.data.image, 
                                           (best_model.coord.icrs.ra.to_value('deg'),
                                            best_model.coord.icrs.dec.to_value('deg')), 
                                           stddev=rms_prof, header=self.data.header, 
                                           binsize=binsize_prof.to_value('deg'))
        r2d = data_yprof_tmp[0]*60
        data_yprof = data_yprof_tmp[1]

        mc_y_data = np.zeros((Nmc, len(r2d)))
        for i in range(Nmc):
            mc_y_data[i,:] = radial_profile_sb(noise_mc[i,:,:], 
                                               (best_model.coord.icrs.ra.to_value('deg'),
                                                best_model.coord.icrs.dec.to_value('deg')), 
                                               stddev=rms_prof, header=self.data.header, 
                                               binsize=binsize_prof.to_value('deg'))[1]
        data_yprof_err = np.std(mc_y_data, axis=0)
        
        #========== Compute best fit observables
        self.setpar_profile(popt, parinfo)
        best_ymap_sph = self.get_radial_model()

        best_y_profile = radial_profile_sb(best_ymap_sph, 
                                           (best_model.coord.icrs.ra.to_value('deg'),
                                            best_model.coord.icrs.dec.to_value('deg')), 
                                           stddev=best_ymap_sph*0+1, header=self.data.header, 
                                           binsize=binsize_prof.to_value('deg'))[1]

        r3d, best_pressure_profile = best_model.get_pressure_profile(np.logspace(1, 4, 100)*u.kpc)
        r3d = r3d.to_value('kpc')
        best_pressure_profile = best_pressure_profile.to_value('keV cm-3')
        
        #========== MC resampling
        MC_ymap_sph         = np.zeros((self.mcmc_Nresamp, best_ymap_sph.shape[0], best_ymap_sph.shape[1]))
        MC_y_profile        = np.zeros((self.mcmc_Nresamp, len(r2d)))
        MC_pressure_profile = np.zeros((self.mcmc_Nresamp, len(r3d)))
        
        MC_pars = np.zeros((self.mcmc_Nresamp, len(popt)))
        isamp = 0
        while isamp < self.mcmc_Nresamp:
            param = np.random.multivariate_normal(popt, pcov)
            cond = np.isfinite(self.prior_profile(param, parinfo)) # make sure params are within limits
            if cond:
                MC_pars[isamp,:] = param
                isamp += 1
                
        for i in range(self.mcmc_Nresamp):
            # Get MC model
            self.setpar_profile(MC_pars[i,:], parinfo)
            MC_ymap_sph[i,:,:] = self.get_radial_model()
            
            # Get MC y profile
            MC_y_profile[i,:] = radial_profile_sb(MC_ymap_sph[i,:,:], 
                                                  (best_model.coord.icrs.ra.to_value('deg'),
                                                   best_model.coord.icrs.dec.to_value('deg')), 
                                                  stddev=self.data.image*0+1, header=self.data.header, 
                                                  binsize=binsize_prof.to_value('deg'))[1]
            # Get MC pressure profile
            MC_pressure_profile[i,:] = self.model.get_pressure_profile(r3d*u.kpc)[1].to_value('keV cm-3')
        
        #========== plots map
        utils_plot.show_fit_result_ymap(self.output_dir+'/CurveFit_Profile_results_y_map.pdf',
                                        self.data.image,
                                        self.data.header,
                                        noise_mc,
                                        best_ymap_sph,
                                        mask=self.data.mask,
                                        visu_smooth=visu_smooth.to_value('arcsec'))
        
        #========== plots ymap profile
        utils_plot.show_fit_ycompton_profile(self.output_dir+'/CurveFit_Profile_results_y_profile.pdf',
                                             r2d, data_yprof, data_yprof_err,
                                             best_y_profile, MC_y_profile,
                                             true_compton_profile=true_compton_profile)
        
        #========== plots pressure profile
        utils_plot.show_fit_result_pressure_profile(self.output_dir+'/CurveFit_Profile_results_P_profile.pdf',
                                                    r3d, best_pressure_profile, MC_pressure_profile,
                                                    true_pressure_profile=true_pressure_profile)
        
        
